"""
======================================
Design of Experiments (:mod:`mqr.doe`)
======================================

.. currentmodule:: mqr.doe

User guide
    :doc:`/user_guide/design-of-experiments`

Detailed examples
    https://github.com/nklsxn/mqrpy-guide

.. rubric:: Classes
.. autosummary::
    :toctree: generated/

    Design
    Transform
    Affine

"""

from dataclasses import dataclass, field
from collections.abc import Iterable
import pyDOE3
import numpy as np
import pandas as pd

@dataclass
class Design:
    """
    An experimental design.

    Designs should normally be constructed using the from_* methods, which wrap
    calls to the pyDOE3 library. Designs can also be constructed manually,
    either from pyDOE3 or any other method, including directly from numpy arrays.
    See :doc:`/user_guide/design-of-experiments`.

    Designs are composable by concatenation with the `+` operator.

    Attributes
    ----------
    names : list[str]
        Names of variables.
    levels : np.ndarray
        Two-dimensional array containing the levels for each experiment, with a
        column for each variables and a row for each run.
    runs : np.ndarray
        Numerical labels for each run. Useful for tracking runs after randomisation.
    pttypes : array_like
        Numerical label for each point type:
        | 0 : centre point
        | 1 : corner point
        | 2 : axial point

    Examples
    --------
    >>> Design.from_full_fact(['x1', 'x2', 'x3'], [2, 2, 2])
       PtType    x1   x2   x3
    1       1  -1.0 -1.0 -1.0
    2       1   1.0 -1.0 -1.0
    3       1  -1.0  1.0 -1.0
    4       1   1.0  1.0 -1.0
    5       1  -1.0 -1.0  1.0
    6       1   1.0 -1.0  1.0
    7       1  -1.0  1.0  1.0
    8       1   1.0  1.0  1.0

    >>> d1 = Design.from_fracfact(['x1', 'x2', 'x3', 'x4'], 'a b c abc')
    >>> d2 = Design.from_centrepoints(['x1', 'x2', 'x3', 'x4'], 3)
    >>> d1.as_block(1) + d2.as_block(2)
        PtType  Block   x1   x2   x3   x4
    1        1      1 -1.0 -1.0 -1.0 -1.0
    2        1      1  1.0 -1.0 -1.0  1.0
    3        1      1 -1.0  1.0 -1.0  1.0
    4        1      1  1.0  1.0 -1.0 -1.0
    5        1      1 -1.0 -1.0  1.0  1.0
    6        1      1  1.0 -1.0  1.0 -1.0
    7        1      1 -1.0  1.0  1.0 -1.0
    8        1      1  1.0  1.0  1.0  1.0
    9        0      2  0.0  0.0  0.0  0.0
    10       0      2  0.0  0.0  0.0  0.0
    11       0      2  0.0  0.0  0.0  0.0
    """
    names: list[str]
    levels: pd.DataFrame
    runs: pd.Index
    pttypes: pd.Series
    blocks: pd.DataFrame

    def replicate(self, n, label=None):
        """
        Create a new design with each run replicated `n` times.

        Parameters
        ----------
        n : int
            The number of replicates to create.
        label : str, optional
            Add a set of blocks with this name, labelling replicates by number.

        Notes
        -----
        Resets the `runs` indexing counting from 1.

        Returns
        -------
        :class:`Design`
            A new design that is a replicated version of this one.
        """
        idx = self.runs.repeat(n)
        new_runs = pd.RangeIndex(1, len(self) * n + 1)
        new_levels = self.levels.loc[idx].set_index(new_runs)
        if self.pttypes is not None:
            new_pttypes = self.pttypes.loc[idx].set_axis(new_runs)
        else:
            new_pttypes = None
        new_blocks = self.blocks.loc[idx].set_index(new_runs)
        if label is not None:
            new_blocks[label] = np.tile(np.arange(n), len(self)) + 1
        return Design(
            names=self.names,
            levels=new_levels,
            runs=new_runs,
            pttypes=new_pttypes,
            blocks=new_blocks)

    def as_block(self, level, name='Block'):
        """
        Return the same set of runs with a block label set to `level`.

        Parameters
        ----------
        level : int, str
            The label/level of this block.
        name : str
            The name of this collection of blocks. Useful if more than one level
            of blocking is required (ie. nested blocks).
        """
        new_blocks = self.blocks.copy()
        new_blocks[name] = level
        return Design(
            names=self.names,
            levels=self.levels,
            runs=self.runs,
            pttypes=self.pttypes,
            blocks=new_blocks)

    def to_df(self):
        """
        Construct a dataframe representation of the design.

        The resulting DataFrame is indexed with runs, has a column for point types
        if they have been defined, has columns for blocks if they are defined,
        and then columns for the inputs levels corresponding to each run/row.

        Returns
        -------
        pandas.DataFrame
            The design.
        """

        df = pd.DataFrame(index=self.runs)
        if self.pttypes is not None:
            df['PtType'] = self.pttypes
        df[self.blocks.columns] = self.blocks
        for name in self.names:
            df[name] = self.levels[name]
        return df

    def get_factor_df(self, name, ref_levels=0.0):
        """
        Create a dataframe containing all unique levels in this design for a
        variable, and a reference level for all others.

        Parameters
        ----------
        name : str
            The factor to isolate.
        ref_levels : float, optional
            The reference level assigned to all other variables.

        Returns
        -------
        pandas.DataFrame
            A dataframe with levels in `name` as rows and variable names as columns.
        """
        df = pd.DataFrame(columns=self.names, dtype=np.float64)
        df[name] = np.sort(np.unique(self.levels[name]))
        df.fillna(ref_levels, inplace=True)
        return df

    def randomise_runs(self, *order):
        """
        Return the same set of runs, randomised over their run labels.

        Parameters
        ----------
        order : {'PtType', block_name, factor_name}, optional
            Keep the runs ordered by this group(s). When more than one `order`
            is specified, the group labelled by the first element is sorted,
            then the group labelled by the second element is sorted within the
            groups of the first, and so on. The default `None` fully randomises
            the runs.

        Returns
        -------
        :class:`Design`
            A copy of this design, randomised.
        """
        rnd = np.random.choice(
            a=self.runs,
            size=len(self.runs),
            replace=False)
        df = self.to_df().loc[rnd]
        df.sort_values(list(order), inplace=True, kind='stable')
        new_order = df.index
        new_levels = self.levels.loc[new_order]
        if self.pttypes is not None:
            new_pttypes = self.pttypes.loc[new_order]
        else:
            new_pttypes = None
        new_blocks = self.blocks.loc[new_order]
        return Design(
            names=self.names,
            levels=new_levels,
            runs=new_order,
            pttypes=new_pttypes,
            blocks=new_blocks)

    def __add__(self, other):
        """
        Concatenate the runs of another design at the end of this design.

        Parameters
        ----------
        other : :class:`Design`
            The design to concatenate.

        Returns
        -------
        :class:`Design`
            This design and `other` concatenated into one, with run labels of
            `other` offset to continue from the end of this design.
        """
        if self.names != other.names:
            raise ValueError('Designs must contain the same variables.')

        new_runs = self.runs.append(other.runs + self.runs.max())
        if (self.pttypes is None) and (other.pttypes is None):
            new_pttypes = None
        elif (self.pttypes is not None) and (other.pttypes is None):
            new_pttypes = pd.concat([
                self.pttypes,
                pd.Series(np.nan, index=other.runs),
            ], axis=0).set_axis(new_runs)
        elif (self.pttypes is None) and (other.pttypes is not None):
            new_pttypes = pd.concat([
                pd.Series(np.nan, index=self.runs),
                other.pttypes
            ], axis=0).set_axis(new_runs)
        else:
            new_pttypes = pd.concat([self.pttypes, other.pttypes], axis=0).set_axis(new_runs)
        new_levels = pd.concat([self.levels, other.levels], axis=0)
        new_levels.set_index(new_runs, inplace=True)
        new_blocks = pd.concat([self.blocks, other.blocks], axis=0)
        new_blocks.set_index(new_runs, inplace=True)

        return Design(
            names=self.names,
            levels=new_levels,
            runs=new_runs,
            pttypes=new_pttypes,
            blocks=new_blocks)

    def __len__(self):
        return len(self.runs)

    def transform(self, **transforms):
        """
        Apply transforms to the levels of this design.

        Parameters
        ----------
        transforms : {`Transform`, callable, dict}
            Keyword args with keywords corresponding to factor names of the design.
            Not all factors need to be listed/transformed.

            | :class:`Transform`
            |   Apply the given transform to the vector of levels.
            | callable
            |   Apply the given function to the vector of levels.
            | dict
            |   Use each level as a key in this dictionary and replace the level
            |   with the returned value. Each level must be a key in the dictionary.

        Returns
        -------
        :class:`Design`
            A copy of this design with new levels.
        """
        new_levels = self.levels.copy()
        for name, tx in transforms.items():
            if callable(tx):
                new_levels[name] = tx(self.levels[name])
            elif isinstance(tx, dict):
                new_levels[name] = np.vectorize(tx.get)(self.levels[name])
            else:
                raise ValueError(f'Unknown transform {tx}.')
        return Design(
            names=self.names,
            levels=new_levels,
            runs=self.runs,
            pttypes=self.pttypes,
            blocks=self.blocks)

    @staticmethod
    def from_levels(names, levels, runs=None):
        """
        Construct a design from an array of levels.

        Parameters
        ----------
        names : list[str]
            List of variable names.
        levels : array_like
            Two-dimensional array of levels, with runs in rows and variables in
            columns.
        runs : array_like, optional
            Array of labels for runs. (Default `None` results in labels counting
            from 1.)

        Returns
        -------
        :class:`Design`
        """
        m, n = levels.shape
        if n != len(names):
            raise AttributeError('Length of `names` must match number of columns in `levels`.')
        if (runs is not None) and len(runs) != m:
            raise AttributeError('Length of `runs` must match number of rows in `levels`.')

        runs = pd.RangeIndex(1, m+1) if (runs is None) else pd.Index(runs)
        levels = pd.DataFrame(levels, index=runs, columns=names)
        blocks = pd.DataFrame(index=runs)
        return Design(
            names=names,
            levels=levels,
            runs=runs,
            pttypes=None,
            blocks=blocks)

    @staticmethod
    def from_fullfact(names, levels, scale_origin=True, pttypes=True):
        """
        Construct a design from `pyDOE3.fullfact(...)`.

        Parameters
        ----------
        names : list[str]
            List of variable names.
        levels : list[int]
            A list of counts of levels, passed directly to `pyDOE3.fullfact(...)`.
        scale_origin : bool, optional
            | `True`
            |   Scales the levels to have integer values centred at the origin.
            | `False`
            |   Levels are set directly from the `pyDOE3` output.

        Notes
        -----
        The point type column is only added when all factors have two levels.

        Returns
        -------
        :class:`Design`
        """
        coded_levels = pyDOE3.fullfact(levels)
        design = Design.from_levels(names, coded_levels)
        if scale_origin:
            value_counts = [len(np.unique(design.levels[name])) for name in design.names]
            mapper = lambda x: Design._scale(len(np.unique(x)))(x)
            design.levels = design.levels.apply(mapper)
        if np.all(np.isclose(levels, 2)) and pttypes:
            design.pttypes = pd.Series(np.ones(len(design), dtype='u1'), design.runs)

        return design

    @staticmethod
    def from_fracfact(names, gen):
        """
        Construct a design from `pyDOE3.fracfact(...)`.

        Parameters
        ----------
        names : list[str]
            List of variable names.
        gen : str
            Yates-labelled generators for each variable, separated by spaces.
            Passed directly to `pyDOE3.fracfact(...)`.

        Returns
        -------
        :class:`Design`
        """
        levels = pyDOE3.fracfact(gen)
        design = Design.from_levels(names, levels)
        design.pttypes = pd.Series(np.ones(len(design), dtype='u1'), design.runs)
        return design

    @staticmethod
    def from_ccdesign(names, center=(0, 0), alpha='orthogonal', face='circumscribed'):
        """
        Construct a design from `pyDOE3.ccdesign(...)`.

        Parameters
        ----------
        names : list[str]
            List of variable names.
        center : tuple, optional
            Passed to `pyDOE3.ccdesign(...)`.
        alpha : str, optional
            Passed to `pyDOE3.ccdesign(...)`.
        face : str, optional
            Passed to `pyDOE3.ccdesign(...)`.

        Returns
        -------
        :class:`Design`
        """
        levels = pyDOE3.ccdesign(len(names), center=center, alpha=alpha, face=face)
        design = Design.from_levels(names, levels)
        design.pttypes = design.levels.apply(Design._pttype, axis=1).astype('u1')
        return design

    @staticmethod
    def from_centrepoints(names, n):
        """
        Construct a design from runs of centrepoints.

        Parameters
        ----------
        names : list[str]
            List of variable names.
        n : int
            Count of runs.

        Returns
        -------
        :class:`Design`
        """
        levels = np.zeros([n, len(names)])
        design = Design.from_levels(names, levels)
        design.pttypes = pd.Series(np.zeros(len(design), dtype='u1'), design.runs)
        return design

    @staticmethod
    def from_axial(names, exclude=None, magnitude=None):
        """
        Construct a design from runs of axial points.

        Parameters
        ----------
        names : list[str]
            List of variable names.
        exclude : list[str] or set[str], optional
            Iterable of names to exclude from construction (the columns still
            exist, but no runs are added).
        magnitude : float, optional
            Magnitude of axial points. Default is sqrt(len(names)).

        Returns
        -------
        :class:`Design`
        """
        if exclude is None:
            exclude = {}

        if magnitude is None:
            magnitude = np.sqrt(len(names))

        n = len(names)
        n_total = n-len(exclude)
        levels = np.zeros([2*n_total, n])
        j = 0
        for i, name in enumerate(names):
            if name not in exclude:
                levels[2*j, i] = -magnitude
                levels[2*j+1, i] = magnitude
                j += 1
        design = Design.from_levels(names, levels)
        design.pttypes = pd.Series(2*np.ones(len(design), dtype='u1'), design.runs)
        return design

    @staticmethod
    def _is_centre(point):
        # Origin is a centre point
        return np.all(np.isclose(point, 0.0))

    @staticmethod
    def _is_corner(point):
        # Same non-zero distance from the origin on all axes
        d = np.abs(point)
        return (d.iloc[0] > 0) and np.all(np.isclose(d.iloc[1:], d.iloc[0]))

    @staticmethod
    def _is_axial(point):
        # One non-zero entry (on an axis)
        return np.sum(~np.isclose(point, 0.0)) == 1

    @staticmethod
    def _pttype(point):
        '''
        NB: only works on non-transformed points. That is:
            - the origin is a centre point
            - any point that is the same distance from the origin on all axes is a corner point
            - any point with only one non-zero entry is an axial point
            - all other points are not classified
        '''
        if Design._is_centre(point):
            return 0
        elif Design._is_corner(point):
            return 1
        elif Design._is_axial(point):
            return 2
        else:
            return None

    @staticmethod
    def _scale(level_count):
        def scale(levels):
            s = 1 if level_count % 2 else 2
            scaled = levels * s
            return scaled - scaled.max() / 2
        return scale

    def _repr_(self):
        return self.to_df()

    def _repr_html_(self):
        return self.to_df().to_html()

@dataclass
class Transform:
    @staticmethod
    def from_map(map):
        """
        Construct an affine transform.

        See :doc:`/user_guide/design-of-experiments` for examples.

        Parameters
        ----------
        map : dict[float, float]
            A dictionary that maps from an existing level to a new level.
            The dict has two float keys, corresponding to existing levels and
            mapping to a float that is the new corresponding level. The two
            pairs exactly define an affine Transform. For example, the dict
            `{0: 10, 1: 20}` transforms `0` to `10` and `1` to `20`. All other
            points will be interpolated/extrapolated along a straight line: `0.5`
            transforms to `15`. As a result, the maps expressed in the dict need
            not correspond to the levels in the current design.

        Returns
        -------
        :class:`Affine`
        """
        [(l, lval), (r, rval)] = map.items()
        scale = (rval - lval) / (r - l)
        translate = lval - l * scale
        return Affine(
            scale=scale,
            translate=translate)

@dataclass
class Affine(Transform):
    """
    An Affine transform for transforming the levels in a Design.

    The scale is applied first, then the translation.
    See :doc:`/user_guide/design-of-experiments` for examples.

    Attributes
    ----------
    scale : float
        Multiplies experiment levels.
    translate : float
        Offsets the experiment levels after they are scaled by `scale`.
    """
    scale: float
    translate: float

    def __call__(self, level):
        """
        Applies this transform to level values.

        Parameters
        ----------
        levels : array_like
            Levels to transform, usually a column from :attr:`Design.levels`.

        Returns
        -------
        array_like
            Transformed levels.
        """
        return level * self.scale + self.translate



from mqr.inference.confint import ConfidenceInterval
from mqr.inference.hyptest import HypothesisTest
from mqr.inference.power import TestPower

import mqr.inference.lib.util as util
import mqr.interop.inference as interop

import numbers
import numpy as np
import scipy
import statsmodels
import warnings

def size_1sample(effect, alpha, beta, alternative='two-sided'):
    """
    Calculate sample size for a t-test for the mean.

    Calls :func:`sm..tt_solve_power <statsmodels.stats.power.tt_solve_power>`.

    Parameters
    ----------
    effect : float
        Required effect size; Cohen's F.
    alpha : float
        Required significance.
    beta : float
        Required beta (1 - power).
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.

    Returns
    -------
    :class:`mqr.inference.power.TestPower`
    """
    alt = interop.alternative(alternative, 'statsmodels')
    power = 1.0 - beta
    warnings.filterwarnings("error")
    try:
        nobs = statsmodels.stats.power.tt_solve_power(
            alpha=alpha,
            power=power,
            effect_size=effect,
            alternative=alt)
    finally:
        warnings.resetwarnings()
    return TestPower(
        name='mean',
        alpha=alpha,
        beta=beta,
        effect=effect,
        alternative=alternative,
        method='t',
        sample_size=nobs)

def size_2sample(effect, alpha, beta, alternative='two-sided'):
    """
    Calculate sample size for test of difference of unpaired means.

    Calls :func:`sm..tt_ind_solve_power <statsmodels.stats.power.tt_ind_solve_power>`.

    Parameters
    ----------
    effect : float
        Required effect size; Cohen's F.
    alpha : float
        Required significance.
    beta : float
        Required beta (1 - power).
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.

    Returns
    -------
    :class:`mqr.inference.power.TestPower`
    """
    alt = interop.alternative(alternative, 'statsmodels')
    power = 1.0 - beta
    warnings.filterwarnings("error")
    try:
        nobs = statsmodels.stats.power.tt_ind_solve_power(
            alpha=alpha,
            power=power,
            effect_size=effect,
            ratio=1.0,
            alternative=alt)
    finally:
        warnings.resetwarnings()
    return TestPower(
        name='difference between means (independent)',
        alpha=alpha,
        beta=beta,
        effect=effect,
        alternative=alternative,
        method='t',
        sample_size=nobs)

def size_paired(effect, alpha, beta, alternative='two-sided'):
    """
    Calculate sample size for test of difference of unpaired means.

    Calls :meth:`sm..solve_power <statsmodels.stats.power.TTestPower.solve_power>`.

    Parameters
    ----------
    effect : float
        Required effect size; Cohen's F.
    alpha : float
        Required significance.
    beta : float
        Required beta (1 - power).
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.

    Returns
    -------
    :class:`mqr.inference.power.TestPower`
    """
    alt = interop.alternative(alternative, 'statsmodels')
    power = 1.0 - beta
    nobs = statsmodels.stats.power.TTestPower().solve_power(
        alpha=alpha,
        power=power,
        effect_size=effect,
        alternative=alt)
    return TestPower(
        name='difference between means (paired)',
        alpha=alpha,
        beta=beta,
        effect=effect,
        alternative=alternative,
        method='t',
        sample_size=nobs)

def confint_1sample(x, conf=0.95, bounded='both', method='t'):
    """
    Confidence interval for mean.

    Parameters
    ----------
    x : array_like
        Interval for the mean of the population from which this data was sampled.
    conf : float, optional
        Confidence level that determines the width of the interval.
    bounded : {'both', 'below', 'above'}, optional
        Which sides of the interval to close.
    method : {'t', 'z'}, optional
        | 't'
        |   Interval from Student's t distribution.
            Calls :meth:`sm..tconfint_mean <statsmodels.stats.weightstats.DescrStatsW.tconfint_mean>`.
        | 'z'
        |   Interval from the z-scores.
            Calls :meth:`sm..zconfint_mean <statsmodels.stats.weightstats.DescrStatsW.zconfint_mean>`.

    Returns
    -------
    :class:`mqr.inference.confint.ConfidenceInterval`
    """
    value = np.mean(x)
    alt = interop.bounded(bounded, 'statsmodels')
    alpha = 1 - conf
    if method == 't':
        lower, upper = statsmodels.stats.api.DescrStatsW(x).tconfint_mean(alpha, alt)
    elif method == 'z':
        lower, upper = statsmodels.stats.api.DescrStatsW(x).zconfint_mean(alpha, alt)
    else:
        raise ValueError(util.method_error_msg(method, ['t', 'z']))

    return ConfidenceInterval(
        name='mean',
        method=method,
        value=value,
        lower=lower,
        upper=upper,
        conf=conf,
        bounded=bounded)

def confint_2sample(x, y, conf=0.95, pooled=True, bounded='both', method='t'):
    """
    Confidence interval for difference of two unpaired means.

    Parameters
    ----------
    x, y : array_like
        Calculate interval for difference between means of these samples.
    conf : float
        Confidence level that determines the width of the interval.
    pooled : bool
        When `True`, the samples have the same variance, `False` otherwise.
    bounded : {'both', 'below', 'above'}, optional
        Which sides of the interval to close.
    method : {'t', 'z'}, optional
        | 't'
        |   Interval based on Student's t distribution.
            Calls :meth:`sm..tconfint_diff <statsmodels.stats.weightstats.CompareMeans.tconfint_diff>`.
        | 'z'
        |   Interval based on z-scores.
            Calls :meth:`sm..zconfint_diff <statsmodels.stats.weightstats.CompareMeans.zconfint_diff>`.

    Returns
    -------
    :class:`mqr.inference.confint.ConfidenceInterval`
    """
    value = np.mean(x) - np.mean(y)
    alt = interop.bounded(bounded, 'statsmodels')
    alpha = 1 - conf
    usevar = 'pooled' if pooled else 'unequal'
    xs = statsmodels.stats.api.DescrStatsW(x)
    ys = statsmodels.stats.api.DescrStatsW(y)
    comp = statsmodels.stats.api.CompareMeans(xs, ys)

    if method == 't':
        lower, upper = comp.tconfint_diff(
            alpha=alpha,
            usevar=usevar,
            alternative=alt)
    elif method == 'z':
        lower, upper = comp.zconfint_diff(
            alpha=alpha,
            usevar=usevar,
            alternative=alt)
    else:
        raise ValueError(util.method_error_msg(method, ['t', 'z']))
    return ConfidenceInterval(
        name='difference between means (independent)',
        method=method,
        value=value,
        lower=lower,
        upper=upper,
        conf=conf,
        bounded=bounded)

def confint_paired(x, y, conf=0.95, bounded='both', method='t'):
    """
    Confidence interval for difference of two paired means.

    Parameters
    ----------
    x, y : array_like
        Calculate interval for difference between means of these samples.
    conf : float, optional
        Confidence level that determines the width of the interval.
    bounded : {'two-sided', 'less', 'greater'}, optional
        Which sides of the interval to close.
    method : {'t', 'z'}, optional
        | 't'
        |   Interval based on Student's t distribution.
            Calls :meth:`sm..tconfint_mean <statsmodels.stats.weightstats.DescrStatsW.tconfint_mean>`.
        | 'z'
        |   Interval based on z-score.
            Calls :meth:`sm..zconfint_mean <statsmodels.stats.weightstats.DescrStatsW.zconfint_mean>`.

    Returns
    -------
    :class:`mqr.inference.confint.ConfidenceInterval`
    """
    delta = x - y
    ci = confint_1sample(delta, conf, bounded, method)
    ci.name = 'difference between means (paired)'
    return ci

def test_1sample(x, H0_mean=0.0, alternative='two-sided', method='t'):
    """
    Hypothesis test for the mean of a sample.

    Null-hypothesis
        mean(`x`) == `H0_mean`

    Parameters
    ----------
    x : array_like
        Test mean of this sample.
    H0_mean : float, optional
        Null-hypothesis mean.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : {'t', 'z'}, optional
        | 't'
        |   Test based on Student's t distribution.
            Calls :func:`scipy..ttest_1samp <scipy.stats.ttest_1samp>`.
        | 'z'
        |   Test based on z-score.
            Calls :func:`sm..ztest <statsmodels.stats.weightstats.ztest>`.

    Returns
    -------
    :class:`mqr.inference.hyptest.HypothesisTest`
    """
    if method == 't':
        statistic, pvalue = scipy.stats.ttest_1samp(
            a=x,
            popmean=H0_mean,
            alternative=alternative)
    elif method == 'z':
        alt = interop.alternative(alternative, lib='statsmodels')
        statistic, pvalue = statsmodels.stats.weightstats.ztest(
            x1=x,
            value=H0_mean,
            alternative=alt)
    else:
        raise ValueError(util.method_error_msg(method, ['t', 'z']))

    x_name = util.var_name(x, 'x')
    return HypothesisTest(
        description='mean',
        alternative=alternative,
        method=method,
        sample_stat=f'mean({x_name})',
        sample_stat_target=H0_mean,
        sample_stat_value=np.mean(x),
        stat=statistic,
        pvalue=pvalue,
    )

def test_2sample(x, y, H0_diff=0.0, pooled=True, alternative='two-sided', method='t'):
    """
    Hypothesis test for the difference between means of two unpaired samples.

    Null-hypothesis
        mean(`x`) - mean(`y`) == `H0_diff`

    Parameters
    ----------
    x, y : array_like
        Test the difference between means of these samples.
    H0_diff : float, optional
        Null-hypothesis difference.
    pooled : bool, optional
        `True` when the samples are taken from the same population, `False` otherwise.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : {'t', 'z'}, optional
        | 't'
        |   Test based on Student's t distribution
            Calls :func:`sm..ttest_ind <statsmodels.stats.weightstats.ttest_ind>`.
        | 'z'
        |   Test based on z-score.
            Calls :func:`sm..ztest <statsmodels.stats.weightstats.ztest>`.

    Returns
    -------
    :class:`mqr.inference.hyptest.HypothesisTest`
    """
    alt = interop.alternative(alternative, 'statsmodels')
    usevar = 'pooled' if pooled else 'unequal'
    if method == 't':
        statistic, pvalue, _dof = statsmodels.stats.weightstats.ttest_ind(
            x1=x,
            x2=y,
            alternative=alt,
            usevar=usevar,
            value=H0_diff)
    elif method == 'z':
        statistic, pvalue = statsmodels.stats.weightstats.ztest(
            x1=x,
            x2=y,
            value=H0_diff,
            alternative=alt,
            usevar=usevar)
    else:
        raise ValueError(util.method_error_msg(method, ['t', 'z']))
    
    x_name = util.var_name(x, 'x')
    y_name = util.var_name(y, 'y')
    return HypothesisTest(
        description='difference between means (independent)',
        alternative=alternative,
        method=method,
        sample_stat=f'mean({x_name}) - mean({y_name})',
        sample_stat_target=H0_diff,
        sample_stat_value=x.mean()-y.mean(),
        stat=statistic,
        pvalue=pvalue,)

def test_paired(x, y, alternative='two-sided', method='t'):
    """
    Hypothesis test for the difference between means of two paired samples.

    Null-hypothesis
        mean(`x`) == mean(`y`)

    Parameters
    ----------
    x, y : array_like
        Test the difference between means of these samples.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : 't', optional
        Only Student's t is currently available.
        Calls :func:`scipy..ttest_rel <scipy.stats.ttest_rel>`.

    Returns
    -------
    :class:`mqr.inference.hyptest.HypothesisTest`
    """
    if method != 't':
        raise ValueError(util.method_error_msg(method, ['t']))

    result = scipy.stats.ttest_rel(x, y, alternative=alternative)

    x_name = util.var_name(x, 'x')
    y_name = util.var_name(y, 'y')
    return HypothesisTest(
        description='difference between means (paired)',
        alternative=alternative,
        method='t',
        sample_stat=f'mean({x_name}) - mean({y_name})',
        sample_stat_target=0,
        sample_stat_value=x.mean()-y.mean(),
        stat=result.statistic,
        pvalue=result.pvalue,)



from mqr.inference.confint import ConfidenceInterval
from mqr.inference.hyptest import HypothesisTest
from mqr.inference.power import TestPower

from mqr.inference.lib import util
import mqr.inference.variance as variance

import numpy as np

def size_1sample(std_ratio, alpha, beta, alternative='two-sided'):
    """
    Calculate sample size for test of standard deviation of a sample.

    Null-hypothesis
        `std_ratio` = sigma / sigma0 == 1

    Parameters
    ----------
    std_ratio : float
        Required effect size.
    alpha : float
        Required significance.
    beta : float
        Required beta (1 - power).
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.

    Returns
    -------
    :class:`mqr.inference.power.TestPower`
    """
    var = variance.size_1sample(np.square(std_ratio), alpha, beta, alternative)
    return TestPower(
        name='standard deviation',
        alpha=var.alpha,
        beta=var.beta,
        effect=std_ratio,
        alternative=alternative,
        method=var.method,
        sample_size=var.sample_size)

def size_2sample(std_ratio, alpha, beta, alternative='two-sided'):
    """
    Calculate sample size for test of ratio of standard deviations.

    Null-hypothesis
        `std_ratio` = sigma_1 / sigma_2 == 1

    Parameters
    ----------
    std_ratio : float
        Required effect size.
    alpha : float
        Required significance.
    beta : float
        Required beta (1 - power).
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.

    Returns
    -------
    :class:`mqr.inference.power.TestPower`
    """
    var = variance.size_2sample(np.square(std_ratio), alpha, beta, alternative)
    return TestPower(
        name='ratio of standard deviations',
        alpha=var.alpha,
        beta=var.beta,
        effect=std_ratio,
        alternative=alternative,
        method=var.method,
        sample_size=var.sample_size)

def confint_1sample(x, conf=0.95, bounded='both', method='chi2'):
    """
    Confidence interval for the standard deviation of a sample.

    Parameters
    ----------
    x : array_like
        Calcaulate interval for the standard deviation of this sample.
    conf : float, optional
        Confidence level that determines the width of the interval.
    bounded : {'both', 'below', 'above'}, optional
        Which sides of the interval to close.
    method : {'chi2'}, optional
        Only an interval based on the 'chi2' distribution is implemented.

    Returns
    -------
    :class:`mqr.inference.confint.ConfidenceInterval`
    """
    var = variance.confint_1sample(x, conf, bounded=bounded, method=method)
    return ConfidenceInterval(
        name="standard deviation",
        method=method,
        value=np.sqrt(var.value),
        lower=np.sqrt(var.lower),
        upper=np.sqrt(var.upper),
        conf=conf,
        bounded=bounded)

def confint_2sample(x, y, conf=0.95, bounded='both', method='f'):
    """
    Confidence interval for the ratio of standard deviations of two samples.

    Parameters
    ----------
    x, y : array_like
        Calculate interval for the ratio of standard deviations of these two samples.
    conf : float, optional
        Confidence level that determines the width of the interval.
    bounded : {'both', 'below', 'above'}, optional
        Which sides of the interval to close.
    method : {'f'}, optional
        Only 'f' is available, which calculates the standard deviation from an
        interval on variance.

    Returns
    -------
    :class:`mqr.inference.confint.ConfidenceInterval`
    """
    var = variance.confint_2sample(x, y, conf, bounded=bounded, method=method)
    lower = 0.0 if bounded == 'above' else np.sqrt(var.lower)
    upper = np.inf if bounded == 'below' else np.sqrt(var.upper)
    return ConfidenceInterval(
        name="ratio of standard deviations",
        method=method,
        value=np.sqrt(var.value),
        lower=lower,
        upper=upper,
        conf=conf,
        bounded=bounded)

def test_1sample(x, H0_std, alternative='two-sided'):
    """
    Hypothesis test for the varianve of a sample.

    Null hypothesis
        var(`x`) / `H0_var` == 1

    Parameters
    ----------
    x : array_like
        Test variance of this sample.
    H0_std : float
        Null-hypothesis standard deviation.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.

    Returns
    -------
    :class:`mqr.inference.hyptest.HypothesisTest`
    """
    var = variance.test_1sample(x, np.square(H0_std), alternative)
    x_name = util.var_name(x, 'x')
    return HypothesisTest(
        description='standard deviation',
        alternative=alternative,
        method='chi2',
        sample_stat=f'std({x_name})',
        sample_stat_target=H0_std,
        sample_stat_value=np.sqrt(var.sample_stat_value),
        stat=var.stat,
        pvalue=var.pvalue)

def test_2sample(x, y, alternative='two-sided'):
    """
    Hypothesis test for the ratio of variances of two samples.

    Null hypothesis
        var(`x`) / var(`y`) == 1

    Parameters
    ----------
    x, y : array_like
        Test ratio of variances of these two samples.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : {'f', 'bartlett'}, optional
        | 'f'
        |   F-test on variances applied to the squares of sample standard deviations.
        | 'bartlett'
        |   Bartlett's test. Calls :func:`scipy..bartlett <scipy.stats.bartlett>`.

    Returns
    -------
    :class:`mqr.inference.hyptest.HypothesisTest`
    """
    var = variance.test_2sample(x, y, alternative, 'f')
    x_name = util.var_name(x, 'x')
    y_name = util.var_name(y, 'y')
    return HypothesisTest(
        description='ratio of standard deviations',
        alternative=alternative,
        method=var.method,
        sample_stat=f'std({x_name}) / std({y_name})',
        sample_stat_target=np.sqrt(var.sample_stat_target),
        sample_stat_value=np.sqrt(var.sample_stat_value),
        stat=var.stat,
        pvalue=var.pvalue)



import mqr

import numpy as np
import pandas as pd

def test_combine():
    mean = 5
    sigma = 1
    zscore = np.array([[0, 3.5, 0, 0, 0, 2.5, 1, 2.5, 2.5, 0]]).T
    data = pd.DataFrame(mean + zscore * sigma)
    params = mqr.spc.XBarParams(centre=mean, sigma=sigma)
    stat = params.statistic(data)

    rule = mqr.spc.rules.combine(
        np.logical_or,
        mqr.spc.rules.limits(),
        mqr.spc.rules.aofb_nsigma(a=3, b=4, n=2))
    alarms = rule(stat, params)

    assert list(alarms.index) == list(stat.stat.index)
    assert list(alarms) == [
        False,
        True, # Larger than UCL => triggers `rules.limits`
        False,
        False,
        False,
        False,
        False,
        False,
        True, # 3 of the 4 points to here (incl) are greater than 2-sigma => `rules.aofb_nsigma`
        False,
    ]

def test_limits():
    mean = 5
    sigma = 1
    zscore = np.array([[0, 3.5, 0, 0, 0, 2.5, 1, 2.5, 2.5, 0]]).T
    data = pd.DataFrame(mean + zscore * sigma)
    params = mqr.spc.XBarParams(centre=mean, sigma=sigma)
    stat = params.statistic(data)

    rule = mqr.spc.rules.limits()
    alarms = rule(stat, params)

    assert list(alarms.index) == list(stat.stat.index)
    assert list(alarms) == [
        False,
        True, # Larger than UCL => triggers `rules.limits`
        False,
        False,
        False,
        False,
        False,
        False,
        False,
        False,
    ]

def test_aofb_nsigma():
    mean = 5
    sigma = 1
    zscore = np.array([[0, 3.5, 0, 0, 0, 2.5, 1, 2.5, 2.5, 0]]).T
    data = pd.DataFrame(mean + zscore * sigma)
    params = mqr.spc.XBarParams(centre=mean, sigma=sigma)
    stat = params.statistic(data)

    rule = mqr.spc.rules.aofb_nsigma(a=3, b=4, n=2)
    alarms = rule(stat, params)

    assert list(alarms.index) == list(stat.stat.index)
    assert list(alarms) == [
        False,
        False,
        False,
        False,
        False,
        False,
        False,
        False,
        True, # 3 of the 4 points to here (incl) are greater than 2-sigma => `rules.aofb_nsigma`
        False,
    ]

def test_n1side():
    mean = 5
    sigma = 1
    zscore = np.array([[0, 3.5, 0, 0, 0, -2.5, -1, 1, 2.5, -2.5, 0]]).T
    data = pd.DataFrame(mean + zscore * sigma)
    params = mqr.spc.XBarParams(centre=mean, sigma=sigma)
    stat = params.statistic(data)

    rule = mqr.spc.rules.n_1side(n=2)
    alarms = rule(stat, params)

    assert list(alarms.index) == list(stat.stat.index)
    assert list(alarms) == [
        False,
        False,
        False,
        False,
        False,
        False,
        True, # -2.5 and -1 are two on the same side => rules.n_1side
        False,
        True, # 1 and 2.5 are two on the same side => rules.n_1side
        False,
        False,
    ]

def test_ntrending():
    mean = 5
    sigma = 1
    zscore = np.array([[0, 3.5, 0, 0, 0, -2.5, -1, 1, 2.5, -2.5, 0]]).T
    data = pd.DataFrame(mean + zscore * sigma)
    params = mqr.spc.XBarParams(centre=mean, sigma=sigma)
    stat = params.statistic(data)

    rule = mqr.spc.rules.n_trending(n=4)
    alarms = rule(stat, params)

    assert list(alarms.index) == list(stat.stat.index)
    assert list(alarms) == [
        False,
        False,
        False,
        False,
        False,
        False,
        False,
        False,
        True, # -2.5, -1, 1 and 2.5 are trending => rules.n_trending
        False,
        False,
    ]



"""
===================================
SPC utilities (:mod:`mqr.spc.util`)
===================================

.. currentmodule: mqr.spc.util

.. rubric:: Functions

.. autosummary::
    :toctree: generated/

    c4
    d2
    d3
    c4_fn
    d2_integral
    d3_integral
    solve_arl
    solve_h4
    group_consecutives
    alarm_subsets

"""

import importlib
import numpy as np
import pandas as pd
import pickle
import scipy

import mqr

# d2_array = np.full([99], np.nan)
# for n in range(2, 101):
#     d2_array[n-2] = mqr.spc.util.d2_fn(n, epsabs=1e-10)
# with open('d2-table.pkl', 'wb') as f:
#     pickle.dump(d2_array, f)

tables = importlib.resources.files('mqr.tables')

with open(tables/'c4-table.pkl', 'rb') as f:
    c4_table = pickle.load(f)

with open(tables/'d2-table.pkl', 'rb') as f:
    d2_table = pickle.load(f)

with open(tables/'d3-table.pkl', 'rb') as f:
    d3_table = pickle.load(f)

def lookup(index, table):
    """
    Lookup a value in a table by index.

    Parameters
    ----------
    index : int, ndarray, Series
        Index or indices to lookup.
    table : {c4_table, d2_table, d3_table}
        Tabulated data.

    Notes
    -----
    No shift is applied to the index, so ensure the provided index corresponds
    to the desired sample size.

    Returns
    -------
    int, ndarray, Series
        Values from the lookup table, with the same type and dimensions as the
        argument `index`.
    """
    if isinstance(index, pd.Series):
        return index.apply(lambda n: table[n])
    else:
        return table[index]

def c4(n):
    """
    Retrieves tabulated values of the unbiasing constant c4.

    Parameters
    ----------
    n : int, ndarray, Series
        Sample size(s).

    Notes
    -----
    The tabulated values were generated using :func:`c4_fn`.

    Returns
    -------
    int, ndarray, Series
        Values from the c4 table, with the same type and dimensions as the
        argument `index`.
    """
    if np.any(n < 2) or np.any(n > 100):
        raise ValueError('Sample size n must be between 2 and 100.')

    return lookup(n-2, c4_table)

def d2(n):
    """
    Retrieves tabulated values of the unbiasing constant d2.

    Parameters
    ----------
    n : int, ndarray, Series
        Sample size(s).

    Notes
    -----
    The tabulated values were generated using :func:`d2_integral`.

    Returns
    -------
    int, ndarray, Series
        Values from the d2 table, with the same type and dimensions as the
        argument `index`.
    """
    if np.any(n < 2) or np.any(n > 100):
        raise ValueError('Sample size n must be between 2 and 100.')

    return lookup(n-2, d2_table)

def d3(n):
    """
    Retrieves tabulated values of the unbiasing constant d3.

    Parameters
    ----------
    n : int, ndarray, Series
        Sample size(s).

    Notes
    -----
    The tabulated values were generated using :func:`d3_integral`.

    Returns
    -------
    int, ndarray, Series
        Values from the d3 table, with the same type and dimensions as the
        argument `index`.
    """
    if np.any(n < 2) or np.any(n > 100):
        raise ValueError('Sample size n must be between 2 and 100.')

    return lookup(n-2, d3_table)

def c4_fn(n):
    """
    Direct calculation of the c4 unbiasing constant from the gamma function.

    Evaluates:

    .. math::
        c_4(n) = \\frac{ \\Gamma\\left(\\frac{n}{2}\\right)\\sqrt{\\frac{n}{n-1}}}{\\Gamma\\left(\\frac{n-1}{2}\\right)} .

    Parameters
    ----------
    n : int
        Sample size.

    Returns
    -------
    float
        The value of c4 for the given sample size.
    """
    num = scipy.special.gamma(n / 2) * np.sqrt(2 / (n - 1))
    den = scipy.special.gamma((n - 1) / 2)
    return num / den

def f2(n):
    """
    Integrand for the integral defining d2.
    """
    dist = scipy.stats.norm()
    def _f2(x):
        phi_x = dist.cdf(x)
        return 1 - (1 - phi_x)**n - phi_x**n
    return _f2

def f3_tr(n):
    """
    Integrand for the integral in d3 (transformed version).

    The transform is a substitution of variables the rotates the integration
    limits through 45deg (in the x-y plane), removining the dependence of the
    inner integral's limit on the outer integral's variable.
    """
    dist = scipy.stats.norm()
    sqrt2 = np.sqrt(2)

    def _f3_tr(s, t):
        x = (s - t) / sqrt2
        y = (s + t) / sqrt2
        phi_x = dist.cdf(x)
        phi_y = dist.cdf(y)
        return 1 - phi_y**n - (1-phi_x)**n + (phi_y - phi_x)**n
    return _f3_tr

def d2_integral(n, **quad_kws):
    """
    Numerical integration to calculate the d2 unbiasing constant.

    Evaluates:

    .. math::
        d_2(n) = \\int_{-\\infty}^{\\infty} f_2(x; n) dx 

    where

    .. math::
        f_2(x; n) = 1 - (1 - \\Phi(x))^n - \\Phi(x)^n .

    Parameters
    ----------
    n : int
        Sample size.
    **quad_kws
        Keyword args passed as `**quad_kws` to :func:`scipy.integrate.quad`.

    Returns
    -------
    float
    """
    return scipy.integrate.quad(f2(n), -np.inf, np.inf, **quad_kws)[0]

def d3_integral(n, d2_fn=None, **dblquad_kws):
    """
    Numerical integration to calculate the d3 unbiasing constant.

    Evaluates:

    .. math::
        d_3(n) = \\sqrt{ 2\\int_{-\\infty}^{\\infty} \\int_{-\\infty}^{y} f_3(x, y; n) dx dy - d_2^2(n) },

    where

    .. math::
        f_3(x, y; n) = 1 - \\Phi(y)^n - (1 - \\Phi(x))^n + (\\Phi(y) - \\Phi(x))^n.

    Evaluating the integral is slightly quicker with the transform
    (which is a rotation of the plane through 45 degrees)

    .. math::
        x = \\frac{s - t}{\\sqrt{2}}, \\\\
        y = \\frac{s + t}{\\sqrt{2}}

    over :math:`s \\in (-\\infty, \\infty)` and :math:`t \\in [0, \\infty)`.
    The transform's Jacobian has determinant equal to 1, so the integral is
    evaluated as

    .. math::
        d_3(n) = \\sqrt{ 2\\int_{0}^{\\infty} \\int_{-\\infty}^{\\infty}
            f_3\\left(\\frac{s - t}{\\sqrt{2}}, \\frac{s + t}{\\sqrt{2}}; n\\right) dx dy - d_2^2(n) } .

    Parameters
    ----------
    n : int
        Sample size
    d2_fn : Callable[int, float], optional
        Function evaluating the d2 unbiasing constant. Default is
        :func:`mqr.spc.util.d2`.
    **dblquad_kws : dict
        Keyword args passed as `**dblquad_kws` to :func:`scipy.integrate.dblquad`.

    Notes
    -----
    Uses a substitution that removes the limit's dependence on variables.

    Returns
    -------
    float
    """
    integral = scipy.integrate.dblquad(f3_tr(n), 0, np.inf, -np.inf, np.inf, **dblquad_kws)[0]
    d2_val = d2_fn(n) if (d2_fn is not None) else d2(n)
    return np.sqrt(2 * integral - d2_val**2)

def solve_arl(h4, p, lmda, N=20):
    """
    Find the in-control ARL of an MEWMA chart.

    Follows the modified Gauss-Legendre method in [1]_ to evaluate:

    .. math::
        L(\\alpha) = 1 + \\int_0^h L(u)/\\lambda^2 f(u/\\lambda^2 | \\eta(\\alpha)) du

    where

    .. math::
        \\begin{gather}
            h = h_4 \\lambda / (2 - \\lambda),\\text{ and} \\\\
            \\eta(\\alpha) = \\alpha ((1 - \\lambda) / \\lambda)^2.
        \\end{gather}

    The algorithm makes a change of variables
    (:math:`\\alpha \\rightarrow \\alpha^2`), which improves its performance.

    Parameters
    ----------
    h4 : float
        Upper control limit on MEWMA chart.
    p : int
        Dimension of monitored vector.
    lmda : float
        Decay factor for EWMA statistic.
    N : int, optional
        Number of quadrature sample points (Gauss-Legendre).

    Returns
    -------
    float

    References
    ----------
    .. [1]  Knoth, S. (2017).
            ARL numerics for MEWMA charts.
            Journal of Quality Technology, 49(1), 78-89.
    """
    a, b = 0, np.sqrt(h4 * lmda / (2 - lmda))
    scale = (b - a) / 2
    shift = (a + b) / 2

    z, w = scipy.special.roots_legendre(N)
    z = z * scale + shift
    w = w * scale

    def eta(alpha):
        return alpha * ((1 - lmda) / lmda)**2

    def fn_W(s, t):
        return scipy.stats.ncx2(p, eta(s**2)).pdf(t**2 / lmda**2) * 2 * t

    def fn_g(s):
        return 1

    c = 1 / lmda**2

    return mqr.utils.fredholm2(0, fn_W, fn_g, c, z, w)

def solve_h4(arl_0, p, lmda, init_h4=15.0):
    """
    Find the UCL of an MEWMA chart from the in-control ARL.

    Parameters
    ----------
    arl_0 : float
        In-control average run length.
    p : int
        Dimension of monitored vector.
    lmda : float
        Decay factor for EWMA statistic.
    init_h4 : float, optional
        Initial guess for the upper limit.

    Returns
    -------
    scipy.optimize.RootResults
        The result from the root finder. Use `result.converged` to check if a
        result was found, and call `result.root` to get the value of the limit, h4.
        See :func:`scipy.optimize.root_scalar` for more info.
    """
    fn = lambda x: solve_arl(x, p, lmda) - arl_0
    result = scipy.optimize.root_scalar(fn, x0=init_h4)
    return result.root, result

def group_consecutives(array, is_consecutive=None):
    """
    Groups the index of an array to form sets of consecutive values.

    Parameters
    ----------
    array : array_like[T]
        Values with possible consecutive elements.
    is_consecutive : Callable[[T, T], bool]
        A function returning True if its two arguments are consecutive.

    Returns
    -------
    list[T]

    Examples
    --------

    >>> consecutive_rule = lambda a, b: b == a + 1
    >>> mqr.spc.util.group_consecutives(
            np.array([1, 2, 4, 6, 7, 8, 10]),
            consecutive_rule)
    [array([1, 2]), array([4]), array([6, 7, 8]), array([10])]

    """
    if len(array) == 0:
        return []
    if is_consecutive is None:
        is_consecutive = lambda a, b: (b - a == 1)
    groups = []
    acc = [array[0]]
    last = array[0]
    for i in array[1:]:
        if not is_consecutive(last, i):
            groups.append(acc)
            acc = []
        acc.append(i)
        last = i
    groups.append(acc)
    return groups

def alarm_subsets(alarms):
    """
    Group alarms into subsets where indices are consecutive.

    Parameters
    ----------
    alarms : pandas.Series[bool]
        Series of control alarm points.

    Returns
    -------
    list[pandas.Series[bool]]
        Lists of alarms with consecutive indices.
    """
    idx = np.where(alarms)[0]
    groups = group_consecutives(idx)
    return [pd.Index(alarms.index[g]) for g in groups]



from dataclasses import dataclass
import numpy as np

import matplotlib as mpl
import matplotlib.patheffects as pe
from cycler import cycler

@dataclass
class Defaults:
    """
    Default values for plotting libraries.

    These defaults are loaded automatically each time the `mqr.plot.Figure`
    context is entered.

    Attributes
    ----------
    rc_params : dict
        Paramters passed to `matplotlib` as defaults for all plots.
    marker : str
        The marker to use when a line plot is drawn with markers.
    line_overlay_path_effects : list[:class:`matplotlib.patheffects.AbstractPathEffect`]
        Path effects that are used to foreground a line on busy or dense plots.
    ishikawa : dict
        Default settings for Ishikawa diagrams. See :func:`mqr.plot.ishikawa`.

    """
    rc_params = {
        'axes.formatter.useoffset': False,
        'axes.prop_cycle': cycler(color=[
            '#4e79a7', # blue
            '#f28e2b', # orange
            '#59a14f', # green
            '#e15759', # red
            '#b070ac', # purple
            '#9c755f', # brown
            '#ff9da7', # pink/salmon
            '#bab0ac', # grey
            '#edc948', # yellow
            '#76b7b2', # turquoise
        ]),
        'lines.linewidth': 1.2,
    }

    # General
    marker = 'o'
    line_overlay_path_effects = [
            pe.Stroke(linewidth=3, alpha=0.5, foreground='w'),
            pe.Normal()]

    # Ishikawa diagrams
    ishikawa = dict(
        head_space=2.0,
        bone_space=8.0, # NB: niklsxn: should be auto based on widest text
        bone_angle=np.pi/4,
        cause_space=1.0,
        cause_length=2.0,
        padding=(1.0, 1.0, 1.0, 1.0),
        line_kwargs=dict(linewidth=0.8, color='k'),
        defect_font_dict=dict(fontsize=12, weight='bold'),
        cause_font_dict=dict(fontsize=9, weight='bold'),
        primary_font_dict=dict(fontsize=9),
    )



from collections.abc import Iterable
import numpy as np

def make_ordinal(n):
    '''
    Convert an integer into its ordinal representation

    Thanks to: https://stackoverflow.com/a/50992575

    Parameters
    ----------
    n : int
        Number to express in ordinal representation
    '''
    if not np.isclose(n, int(n)):
        number = str(n)
        suffix = 'th'
    elif (11 <= (n % 100) <= 13) or ():
        number = str(int(n))
        suffix = 'th'
    else:
        number = str(int(n))
        suffix = ['th', 'st', 'nd', 'rd', 'th'][min(int(n) % 10, 4)]
    return number + suffix

def clip_where(a, a_min, a_max, where):
    '''
    Clamp `a` so that `a_min <= a <= a_max` is true or all true, masked by `where`.

    Parameters
    ----------
    a : number or array_like
        Value or values to clip.
    a_min : number
        Minimum value.
    a_max : number
        Maximum value.
    where : bool or array[bool]
        Clip when this value or corresponding element evaluates to `True`.
    '''
    aa = np.atleast_1d(a).copy()
    if not isinstance(a_min, Iterable):
        a_min = np.full(aa.size, a_min)
    if not isinstance(a_max, Iterable):
        a_max = np.full(aa.size, a_max)
    if not isinstance(where, Iterable):
        where = np.full(aa.size, where)

    if len(aa) != len(where):
        raise ValueError('Lengths of `a` and `where` must be equal.')
    if len(aa) != len(a_min):
        raise ValueError('Lengths of `a` and `a_min` must be equal.')
    if len(aa) != len(a_max):
        raise ValueError('Lengths of `a` and `a_max` must be equal.')

    aa[np.where(where)] = np.clip(aa[np.where(where)], a_min[np.where(where)], a_max[np.where(where)])
    if isinstance(a, Iterable):
        return aa
    else:
        return aa[0]

def fredholm2(t0, fn_K, fn_g, lmda, x, w):
    """
    Solve a Fredholm equation of the second kind.

    Evaluates f at t0 in:

    .. math::

        f(t) = g(t) + λ \\int_Q K(t, s) f(s) ds

    where the definite integral is approximated over the region Q defined by the
    quadrature points `x` and weights `w`.

    Uses the quadrature points and weights provided to form linear equations,
    and then interpolates the result using Nystrom's method. See [1]_ for a summary.

    Parameters
    ----------
    t0 : float
        Point to evaluate the function `f(t)`.
    fn_K : Callable[]
        Function that returns values of `K(t, s)`; see equation above.
    fn_g : Callable[]
        Function that returns values of `g(t)`; see equation above.
    lmbda : float
        Multiple of the integral term; see equation above.
    x : array_like
        Quadrature abcissa points.
    w : array_like
        Quadrature weights.

    References
    ----------
    .. [1]  Press, W. H. (2007).
            Numerical recipes 3rd edition: The art of scientific computing.
            Cambridge university press.
    """

    if len(x) != len(w):
        raise ValueError('Vectors x and w must be the same length.')

    N = len(x)
    g = np.full([N, 1], np.nan)
    K = np.full([N, N], np.nan)
    I = np.eye(N)

    for i in range(N):
        g[i] = fn_g(x[i])
        for j in range(N):
            K[i, j] = w[j] * fn_K(x[i], x[j])

    L = np.linalg.solve(I - lmda * K, g)[:, 0]

    # Nystrom's interpolation
    return fn_g(t0) + lmda * np.sum(w * L * fn_K(t0, x))



"""
=======================
Plots (:mod:`mqr.plot`)
=======================

.. currentmodule:: mqr.plot

User guide
    :doc:`/user_guide/plots`

Detailed examples
    https://github.com/nklsxn/mqrpy-guide


.. rubric:: Functions
.. autosummary::
    :toctree: generated/

    Figure
    ishikawa
    confint

.. rubric:: Modules
.. autosummary::

    process
    correlation
    regression
    anova
    msa
    spc
    tools

.. rubric:: Modules
.. autosummary::

    tools

.. toctree::
    :maxdepth: 1
    :hidden:
    :titlesonly:

    plot/mqr.plot.process
    plot/mqr.plot.correlation
    plot/mqr.plot.regression
    plot/mqr.plot.anova
    plot/mqr.plot.msa
    plot/mqr.plot.spc
    plot/mqr.plot.tools
"""
from mqr.plot.figure import Figure

from mqr.plot.confint import confint
from mqr.plot.ishikawa import ishikawa
from mqr.plot.lib.util import grouped_df

import mqr.plot.process as process
import mqr.plot.correlation as correlation
import mqr.plot.regression as regression
import mqr.plot.anova as anova
import mqr.plot.msa as msa
import mqr.plot.spc as spc
import mqr.plot.tools as tools

from mqr.plot.lib.util import grouped_df



def bounded_error_msg(bounded):
    return f'Invalid bound "{bounded}". Use both, below or above.'

def compare_error_msg(compare):
    return f'Invalid comparison "{compare}". Use diff or ratio.'

def method_error_msg(method, available):
    msg = f'Method "{method}" is not available. '
    if len(available) == 1:
        msg += f'Use {available[0]}.'
    else:
        msg += f'Use {", ".join(available[:-1])} or {available[-1]}.'
    return msg

def alternative_error_msg(alternative):
    return f'Invalid alternative "{alternative}". Use two-sided, less or greater.'

def var_name(var, default):
    if hasattr(var, 'name') and (var.name is not None) and (len(var.name) > 0):
        return var.name
    else:
        return default



import mqr.inference.lib.util as util

def alternative(alt, lib):
    """
    Convert an alternative string from the scipy convention to another convention.

    Parameters
    ----------
    alt : {'two-sided', 'greater', 'less'}
        Sense of alternative hypothesis to convert.
    lib : {'statsmodels'}
        Library whose convention to convert to. Only 'statsmodels' currently used.

    Returns
    -------
    str
        Sense of alternative expressed in convention of ``lib``.
    """
    if lib == 'statsmodels':
        if alt == 'two-sided':
            return 'two-sided'
        elif alt == 'less':
            return 'smaller'
        elif alt == 'greater':
            return 'larger'
        else:
            raise ValueError(util.alternative_error_msg(alt))
    raise ValueError(f'Invalid library {lib}.')

def bounded(bounded, lib):
    """
    Convert bounds argument of a confidence interval to another library's convention.

    Parameters
    ----------
    bounded : {'both', 'above', 'below'}
        Which side of the confidence interval is bounded.
    lib : {'statsmodels', 'scipy'}
        Convert to this library's convention.

    Returns
    -------
    str
        Sense of 'alternative' expressed in convention of `lib`.
    """
    if lib == 'statsmodels':
        if bounded == 'both':
            return 'two-sided'
        elif bounded == 'below':
            return 'larger'
        elif bounded == 'above':
            return 'smaller'
        else:
            raise ValueError(util.bounded_error_msg(bounded))
    elif lib == 'scipy':
        if bounded == 'both':
            return 'two-sided'
        elif bounded == 'below':
            return 'greater'
        elif bounded == 'above':
            return 'less'
        else:
            raise ValueError(util.bounded_error_msg(bounded))
    else:
        raise ValueError(f'Invalid library {lib}.')



import mqr

import numpy as np
import pandas as pd
import pytest
import scipy

@pytest.fixture
def data():
    np.random.seed(0)
    return pd.DataFrame({
        'x': scipy.stats.norm(1, 2).rvs(100),
        'y': scipy.stats.norm(3, 4).rvs(100),
        'z': scipy.stats.norm(5, 6).rvs(100),
    })

def test_zscore(data):
    stats, z, z_inv = mqr.transforms.zscore(data)
    assert list(stats.loc['mean']) == pytest.approx([1, 3, 5], abs=0.5)
    assert list(stats.loc['std']) == pytest.approx([2, 4, 6], abs=0.5)
    assert np.all(np.isclose(z_inv(z(data['x'])), data['x']))
    assert np.all(np.isclose(z(z_inv(data['x'])), data['x']))
    assert np.all(np.isclose(z_inv(z(data)), data))
    assert np.all(np.isclose(z(z_inv(data)), data))

    stats, z, z_inv = mqr.transforms.zscore(data, 3, 4)
    assert np.all(np.isclose(z_inv(z(data)), data))
    assert np.all(np.isclose(z(z_inv(data)), data))



from mqr.inference.hyptest import HypothesisTest
import mqr.inference.lib.util as util

def test_1sample(x, H0_median=0.0, alternative='two-sided', method='sign'):
    """
    Hypothesis test for the median of a sample.

    Null-hypothesis
        median(`x`) == H0_median

    Parameters
    ----------
    x : array_like
        Test the median of this sample.
    H0_median : float, optional
        Null-hypothesis median.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : {'sign', 'wilcoxon'}, optional
        | 'sign'
        |   Sign test.
            Calls :func:`sm..sign_test <statsmodels.stats.descriptivestats.sign_test>`.
        | 'wilcoxon'
        |   Wilcoxon signed-rank test.
            Calls :func:`scipy..wilcoxon <scipy.stats.wilcoxon>`.

    Returns
    -------
    :class:`mqr.inference.hyptest.HypothesisTest`
    """
    import mqr, statsmodels
    import numpy as np, scipy.stats as st

    if method == 'sign':
        if alternative != 'two-sided':
            raise ValueError(util.alternative_error_msg(alternative))
        stat, pvalue = statsmodels.stats.descriptivestats.sign_test(x, mu0=H0_median)
    elif method == 'wilcoxon':
        stat, pvalue = st.wilcoxon(x-H0_median, alternative=alternative)
    else:
        raise ValueError(util.method_error_msg(method, ['sign', 'wilcoxon']))

    x_name = util.var_name(x, 'x')
    return HypothesisTest(
        description='median',
        alternative=alternative,
        method=method,
        sample_stat=f'median({x_name})',
        sample_stat_target=H0_median,
        sample_stat_value=np.median(x),
        stat=stat,
        pvalue=pvalue,
    )

def test_nsample(*x, alternative='two-sided', method='kruskal-wallis'):
    """
    Hypothesis test for equality of medians amongst samples.

    Null-hypothesis
        median(`x[i]`) == median(`x[j]`), for all `i`, `j`

    Parameters
    ----------
    x : array_like
        Test the medians of these samples.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : {'kruskal-wallis', 'mann-whitney'}, optional
        | 'kruskal-wallis'
        |   Kruskal-Wallis test. A one-way ANOVA based on ranks.
            Calls :func:`scipy..kruskal <scipy.stats.kruskal>`.
        | 'mann-whitney'
        |   Mann-Whitney U test for the differences between two samples.
            Calls :func:`scipy..mannwhitneyu <scipy.stats.mannwhitneyu>`.

    Returns
    -------
    :class:`mqr.inference.hyptest.HypothesisTest`
    """
    import mqr, numpy as np, scipy.stats as st

    if method == 'kruskal-wallis':
        if alternative != 'two-sided':
            raise ValueError(util.alternative_error_msg(alternative))
        sample_stat = 'median(x_i)'
        sample_stat_value = 'median(x_j)'
        stat, pvalue = st.kruskal(*x)
    elif method == 'mann-whitney':
        n = len(x)
        if n != 2:
            raise ValueError(f'Mann-Whitney test cannot be applied to {n} samples (compare exactly 2)')
        sample_stat = 'median(x) - median(y)'
        sample_stat_value = 0.0
        stat, pvalue = st.mannwhitneyu(x[0], x[1], alternative=alternative)
    else:
        raise ValueError(util.method_error_msg(method, ['kruskal-wallis', 'mann-whitney']))

    return HypothesisTest(
        description='equality of medians',
        alternative=alternative,
        method=method,
        sample_stat=sample_stat,
        sample_stat_target=sample_stat_value,
        sample_stat_value=np.nan,
        stat=stat,
        pvalue=pvalue,)



"""
Result type and presentation of confidence intervals.
"""

from mqr.inference.lib.util import bounded_error_msg

from dataclasses import dataclass
import numpy as np
import scipy.stats as st
import warnings

import matplotlib.pyplot as plt

@dataclass
class ConfidenceInterval:
    """
    Result type for confidence interval calculations.

    Attributes
    ----------
    name: str
        Statistic on which the confidence interval was calculated.
    method: str
        Statistical method for determining the interval.
    value: float
        Value of the bounded statistic.
    lower: float
        Lower limit of the interval.
    upper: float
        Upper limit of the interval.
    conf: float
        Confidence that determines the width of the interval.
    bounded: {'both', 'below', 'above'}
        Which sides of the statistic are bounded by this interval.

    Notes
    -----
    The type is iterable, and iterates over the lower and upper bounds (in that
    order).

    Examples
    --------
    The results of confidence interval calculations will be return like this.

    >>> import mqr
    >>> ci = mqr.inference.confint.ConfidenceInterval(
    >>>     'name', 'method',
    >>>     value=1.0, lower=0.0, upper=2.0,
    >>>     conf=0.98, bounded='both')
    >>> print(ci)
    ConfidenceInterval(
        name='test', method='method',
        value=1.0, lower=0.0, upper=2.0,
        conf=0.98, bounded='both')

    Iterate over the lower and upper bounds:

    >>> lower, upper = ci
    >>> lower, upper
    (0.0, 2.0)

    When displayed in a jupyter notebook, this class is shown as an HTML table:

    +-----------+-----------+-----------+
    | | **Confidence Interval**         |
    | | name                            |
    | | 98% (fisher-z)                  |
    +===========+===========+===========+
    | **value** | **lower** | **upper** |
    +-----------+-----------+-----------+
    | 1.0       | 0.0       | 2.0       |
    +-----------+-----------+-----------+
    """

    name: str
    method: str
    value: np.float64
    lower: np.float64
    upper: np.float64
    conf: np.float64
    bounded: str

    def __iter__(self):
        """
        Iterator over the lower and upper bounds of the confidence interval.
        """
        return iter((self.lower, self.upper))

    def _html(self):
        alpha = 1 - self.conf
        if self.bounded == 'both':
            left_alpha = f'{alpha*100/2:g}%'
            right_alpha = f'{(1-alpha/2)*100:g}%'
        elif self.bounded == 'above':
            left_alpha = ''
            right_alpha = f'{(1-alpha)*100:g}%'
        elif self.bounded == 'below':
            left_alpha = f'{alpha*100:g}%'
            right_alpha = ''
        else:
            raise ValueError(bounded_error_msg(bounded))

        return f'''
        <table>
        <thead>
            <tr style='padding-bottom:0px;'>
                <th scope="col" colspan=3 style="text-align:left; padding-bottom:0px;">Confidence Interval</th>
            </tr>
            <tr style='padding-top:0px; padding-bottom:0px;'>
                <td colspan=3 style='text-align: left; padding-top:0px; padding-bottom:0px;'>{self.name}</td>
            </tr>
            <tr style='padding-top:0px;'>
                <td style='text-align:left; padding-top:0px' colspan=3>{self.conf*100:g}% ({self.method})</td>
            </tr>
        </thead>
        <tbody>
            <tr>
                <th scope='col' style='text-align: left;'>value</th>
                <th scope='col' style='text-align: left;'>lower</th>
                <th scope='col' style='text-align: left;'>upper</th>
            </tr>
            <tr>
                <td style='text-align: left;'>{self.value:g}</td>
                <td style='text-align: left;'>{self.lower:g}</td>
                <td style='text-align: left;'>{self.upper:g}</td>
            </tr>
        </tbody>
        </table>
        '''

    def _repr_html_(self):
        return self._html()



from mqr.inference.confint import ConfidenceInterval
from mqr.inference.hyptest import HypothesisTest

from mqr.inference.lib import util
import mqr.interop.inference as interop
import mqr.utils

import numpy as np
import scipy

def confint_1sample(x, q=0.5, conf=0.95, bounded='both'):
    """
    Confidence interval for the quantile of a sample.
    
    Calls :func:`scipy <scipy.stats.quantile_test>`.

    Parameters
    ----------
    x : array_like
        Calculate the confidence interval for the quantile of this sample.
    q : float, optional
        Calculate the interval around this quantile.
    conf : float, optional
        Confidence level that determines the width of the interval.
    bounded : {'both', 'below', 'above'}, optional
        Which sides of the interval to close.

    Returns
    -------
    :class:`mqr.inference.confint.ConfidenceInterval`
    """
    value = np.quantile(x, q)
    alt = interop.bounded(bounded, 'scipy')
    res = scipy.stats.quantile_test(x, q=np.quantile(x, q), p=q, alternative=alt)
    ci = res.confidence_interval(conf)
    percentile = mqr.utils.make_ordinal(100*q)
    return ConfidenceInterval(
        name=f'quantile ({percentile} percentile)',
        method='binom',
        value=value,
        lower=ci.low,
        upper=ci.high,
        conf=conf,
        bounded=bounded)

def test_1sample(x, H0_quant=None, q=0.5, alternative='two-sided'):
    """
    Hypothesis test for the quantile of a sample.

    Null-hypothesis
        quantile(`x`, `q`) == `H0_quant`.

    Calls :func:`scipy..quantile_test <scipy.stats.quantile_test>`.

    Parameters
    ----------
    x : array_like
        Test quantile of this sample.
    H0_quant : float, optional
        Null-hypothesis value associated with ``q``. Default is the ``q`` th
        quantile of ``x``.
    q : float, optional
        Test the value associated with this quantile.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.

    Returns
    -------
    :class:`mqr.inference.hyptest.HypothesisTest`
    """
    if H0_quant is None:
        H0_quant = np.quantile(x, q)

    res = scipy.stats.quantile_test(x, q=H0_quant, p=q, alternative=alternative)

    x_name = util.var_name(x, 'x')
    return HypothesisTest(
        description='quantile',
        alternative=alternative,
        method='binom',
        sample_stat=f'quantile({x_name}, {q})',
        sample_stat_target=H0_quant,
        sample_stat_value=np.quantile(x, q),
        stat=res.statistic,
        pvalue=res.pvalue,)



"""
Basic statistical inference (:mod:`mqr.inference.nonparametric`)
================================================================

.. :currentmodule:`mqr.inference.nonparametric`

User guide
    :doc:`/user_guide/inference`

Detailed examples
    https://github.com/nklsxn/mqrpy-guide

.. rubric:: Non-parametric statistics

.. autosummary::
    :template: autosummary/submodule.rst
    :toctree: generated/

    correlation
    dist
    median
    quantile
    variance
"""

import mqr.inference.nonparametric.correlation
import mqr.inference.nonparametric.dist
import mqr.inference.nonparametric.median
import mqr.inference.nonparametric.quantile
import mqr.inference.nonparametric.variance



import mqr.inference.lib.util as util
from mqr.utils import clip_where

import numpy as np
import scipy

import warnings

def confint_1sample_chi2(count, n, meas, conf, bounded):
    """
    Confidence interval for rate `count / n / meas`.

    Uses the chi-squared method, method 1 in [1]_.

    Parameters
    ----------
    count : int
        Number of events.
    n : int
        Number of periods over which events were counted.
    meas : float
        Extent of one period of observation.
    conf : float
        Confidence level that determines the width of the interval.
    bounded : {'both', 'below', 'above'}
        Which sides of the interval to close.

    Returns
    -------
    :class:`mqr.inference.confint.ConfidenceInterval`

    References
    ----------
    .. [1]  Patil, V. V., & Kulkarni, H. V. (2012).
            Comparison of confidence intervals for the Poisson mean: some new aspects.
            REVSTAT-Statistical Journal, 10(2), 211-22.
    """
    alpha = 1 - conf
    if bounded == 'both':
        lower = scipy.stats.chi2(2 * count).ppf(alpha / 2) / (2 * n * meas)
        upper = scipy.stats.chi2(2 * count + 2).ppf(1 - alpha / 2) / (2 * n * meas)
    elif bounded == 'below':
        lower = scipy.stats.chi2(2 * count).ppf(alpha) / (2 * n * meas)
        upper = np.inf
    elif bounded == 'above':
        lower = 0.0
        upper = scipy.stats.chi2(2 * count + 2).ppf(1 - alpha) / (2 * n * meas)
    else:
        raise ValueError(util.bounded_error_msg(bounded))
    return lower, upper

def confint_1sample_exact(count, n, meas, conf, bounded):
    """
    Confidence interval for rate `count / n / meas`.

    Uses the exact method, method 9 in [1]_.

    Parameters
    ----------
    count : int
        Number of events.
    n : int
        Number of periods over which events were counted.
    meas : float
        Extent of one period of observation.
    conf : float
        Confidence level that determines the width of the interval.
    bounded : {'both', 'below', 'above'}
        Which sides of the interval to close.

    Notes
    -----
    Searches the Poisson cumulative distribution for a rate that that produces
    the given confidence when `count` and `n` have the specified values.

    Returns
    -------
    :class:`mqr.inference.confint.ConfidenceInterval`

    References
    ----------
    .. [1]  Barker, L. (2002).
            A comparison of nine confidence intervals for a Poisson parameter when
            the expected number of events is ≤ 5.
            The American Statistician, 56(2), 85-89.
    """
    if (count < 0) or (n < 0) or (meas < 0):
        raise ValueError(f'Arguments `count`, `n` and `meas` must all be non-negative.')
    alpha = (1 - conf) / 2 if (bounded == 'both') else (1 - conf)
    f_ineq_L = lambda mu: alpha - 1 + scipy.stats.poisson(mu*n).cdf(count)
    f_ineq_U = lambda mu: alpha - scipy.stats.poisson(mu*n).cdf(count)
    constraints_L = ({'type': 'ineq', 'fun': f_ineq_L})
    constraints_U = ({'type': 'ineq', 'fun': f_ineq_U})
    min_fun = lambda mu: mu
    max_fun = lambda mu: -mu
    if bounded == 'both':
        res_L = scipy.optimize.minimize(max_fun, count/n, constraints=constraints_L)
        res_U = scipy.optimize.minimize(min_fun, count/n, constraints=constraints_U)
        lower = res_L.x[0] if res_L.success else 0.0
        upper = res_U.x[0]
    elif bounded == 'below':
        res_L = scipy.optimize.minimize(max_fun, count/n, constraints=constraints_L)
        lower = res_L.x[0] if res_L.success else 0.0
        upper = np.inf
    elif bounded == 'above':
        res_U = scipy.optimize.minimize(min_fun, count/n, constraints=constraints_U)
        lower = 0.0
        upper = res_U.x[0]
    else:
        raise ValueError(util.bounded_error_msg(bounded))
    lower, upper = lower / meas, upper / meas
    return lower, upper

def confint_1sample_wald_cc(count, n, meas, conf, bounded):
    """
    Confidence interval for rate `count / n / meas`.

    Uses the Wald method with continuity correction, method 5 in [1]_ and
    method 3 in [2]_.

    Parameters
    ----------
    count : int
        Number of events.
    n : int
        Number of periods over which events were counted.
    meas : float
        Extent of one period of observation.
    conf : float
        Confidence level that determines the width of the interval.
    bounded : {'both', 'below', 'above'}
        Which sides of the interval to close.

    Notes
    -----
    The continuity correction is applied to both mean and variance.
    See [1]_ and [2]_.

    Returns
    -------
    :class:`mqr.inference.confint.ConfidenceInterval`

    References
    ----------
    .. [1]  Patil, V. V., & Kulkarni, H. V. (2012).
            Comparison of confidence intervals for the Poisson mean: some new aspects.
            REVSTAT-Statistical Journal, 10(2), 211-22.
    .. [2]  Barker, L. (2002).
            A comparison of nine confidence intervals for a Poisson parameter when
            the expected number of events is ≤ 5.
            The American Statistician, 56(2), 85-89.
    """
    alpha = 1 - conf
    value = count / n / meas
    dist = scipy.stats.norm()
    if bounded == 'both':
        lower = (count - 0.5 + dist.ppf(alpha / 2) * np.sqrt(count - 0.5)) / (n * meas)
        upper = (count + 0.5 + dist.ppf(1 - alpha / 2) * np.sqrt(count + 0.5)) / (n * meas)
    elif bounded == 'below':
        lower = (count - 0.5 + dist.ppf(alpha) * np.sqrt(count - 0.5)) / (n * meas)
        upper = np.inf
    elif bounded == 'above':
        lower = 0.0
        upper = (count + 0.5 + dist.ppf(1 - alpha) * np.sqrt(count + 0.5)) / (n * meas)
    else:
        raise ValueError(util.bounded_error_msg(bounded))
    lower = np.clip(lower, 0.0, np.inf)
    return lower, upper

def confint_2sample_wald(count1, n1, count2, n2, meas1, meas2, conf, bounded):
    """
    Confidence interval for difference of rates `count1 / n1 / meas1 - count2 / n2 / meas2`.

    Uses Wald method, equation (9) in [1]_.

    Parameters
    ----------
    count1 : int
        Number of events in first observation.
    n1 : int
        Number of periods over which first events were counted.
    count2 : int
        Number of events in second observation.
    n2 : int
        Number of periods over which second events were counted.
    meas1 : float
        Extent of one period in first observation.
    meas2 : float
        Extent of one period in second observation.
    conf : float
        Confidence level that determines the width of the interval.
    bounded : {'both', 'below', 'above'}
        Which sides of the interval to close.

    References
    ----------
    .. [1]  Krishnamoorthy, K., & Lee, M. (2013).
            New approximate confidence intervals for the difference between
            two Poisson means and comparison.
            Journal of Statistical Computation and Simulation, 83(12), 2232-2243.
    """
    alpha = 1 - conf
    r1 = count1 / n1 / meas1
    r2 = count2 / n2 / meas2
    mu = r1 - r2
    sigma = np.sqrt(r1 / n1 / meas1 + r2 / n2 / meas2)
    dist = scipy.stats.norm(mu, sigma)
    if bounded == 'both':
        lower = dist.ppf(alpha / 2)
        upper = dist.ppf(1 - alpha / 2)
    elif bounded == 'below':
        lower = dist.ppf(alpha)
        upper = np.clip(lower, np.inf, np.inf)
    elif bounded == 'above':
        upper = dist.ppf(1 - alpha)
        lower = np.clip(upper, -np.inf, -np.inf)
    else:
        raise ValueError(util.bounded_error_msg(bounded))
    return lower, upper

def confint_2sample_wald_moment(count1, n1, count2, n2, meas1, meas2, conf, bounded):
    """
    Confidence interval for difference of rates `count1 / n1 / meas1 - count2 / n2 / meas2`.

    Uses the method referred to as the "moment CI", equation (8) in [1]_.

    Parameters
    ----------
    count1 : int
        Number of events in first observation.
    n1 : int
        Number of periods over which first events were counted.
    count2 : int
        Number of events in second observation.
    n2 : int
        Number of periods over which second events were counted.
    meas1 : float
        Extent of one period in first observation.
    meas2 : float
        Extent of one period in second observation.
    conf : float
        Confidence level that determines the width of the interval.
    bounded : {'both', 'below', 'above'}
        Which sides of the interval to close.

    References
    ----------
    .. [1]  Krishnamoorthy, K., & Lee, M. (2013).
            New approximate confidence intervals for the difference between
            two Poisson means and comparison.
            Journal of Statistical Computation and Simulation, 83(12), 2232-2243.
    """
    alpha = 1 - conf
    if bounded == 'both':
        z = scipy.stats.norm().ppf(1 - alpha / 2)
    elif (bounded == 'below') or (bounded == 'above'):
        z = scipy.stats.norm().ppf(1 - alpha)
    else:
        raise ValueError(util.bounded_error_msg(bounded))
    r1 = count1 / n1 / meas1
    r2 = count2 / n2 / meas2
    mu_adj = z**2 * (1 / n1 / meas1 - 1 / n2 / meas2) / 2
    var_adj = z**2 * (1 / n1 / meas1 - 1 / n2 / meas2)**2 / 4
    mu = r1 - r2 + mu_adj
    var = r1 / n1 / meas1 + r2 / n2 / meas2 + var_adj
    if bounded == 'both':
        lower = mu - z * np.sqrt(var)
        upper = mu + z * np.sqrt(var)
    elif bounded == 'below':
        lower = mu - z * np.sqrt(var)
        upper = np.clip(lower, np.inf, np.inf)
    elif bounded == 'above':
        upper = mu + z * np.sqrt(var)
        lower = np.clip(upper, -np.inf, -np.inf)
    else:
        raise ValueError(util.bounded_error_msg(bounded))
    return lower, upper



"""
===================================
Notebook tools (:mod:`mqr.nbtools`)
===================================

Tools for presenting results in notebooks.

.. rubric:: Functions

.. autosummary::
    :toctree: generated/

    grab_figure
    hstack
    vstack
"""

from base64 import b64encode
from enum import Enum
import IPython
from IPython.core.pylabtools import print_figure
from IPython.display import display, HTML
import markdown
from matplotlib._pylab_helpers import Gcf

Line = Enum('Line', ['HORIZONTAL', 'VERTICAL'])

def set_container_width(width_pct: int):
    """
    Changes the Jupyter notebook HTML container width.

    Useful on widescreens to use available screen space.

    Parameters
    ----------
    width_pct : int
        Expand the notebook width to this much of the view. Eg. to take up 95%
        of the page, use `width_pct=95`.

    """
    assert width_pct > 0 and width_pct <= 100, f'Argument `width_pct={width_pct}` must be 0<width_pct<=100'
    display(HTML(f'<style>.container {{ width:{width_pct}% !important; }}</style>'))

def grab_figure(figure, suppress=True):
    '''
    Renders the figure to png, base-64 encoded HTML img tag.

    The resulting HTML will be rendered if it is the last expression in a cell.
    It can also be passed to :func:`hstack` or :func:`vstack` to arrange plots
    around other outputs.

    Parameters
    ----------
    figure : matplotlib.figure.Figure
        The figure to capture.
    suppress : bool, optional
        Suppresses the figure by destroying the object. Since this function is
        used to control how the figure is displayed, the default is to destroy
        the object after capturing its output.
    '''
    raw_data_b64 = b64encode(print_figure(figure)).decode('utf-8')
    image_data = f'data:image/png;base64,{raw_data_b64}'
    if suppress:
        Gcf.destroy_fig(figure)
    return HTML(f'<img src="{image_data:s}" />')

def hstack(*args, margin='5px 10px 5px 10px',
        justify_content='stretch', justify_items='center',
        align_content='start', align_items='start'):
    '''
    Horizontally stack the html, markdown or string representation of `args`.

    The `args` will be stacked in order from left to right, and converted like this:

    * if the arg has the attribute `__repr__`, then its output is used pre-formatted,
    * if the arg has the attribute `_repr_html_`, then its output is used,
    * if the arg is a string, then it is treated as markdown and converted to HTML,
    * otherwise, the arg's default string conversion is used.

    Parameters
    ----------
    args
        Elements to stack horizontally.
    margin : float
        Set in the style attribute of a div wrapping elements in this stack
        (excluding lines, which have a fixed 5px margin).
    justify_content : str, optional
        Set in the style attribute of the flexbox containing this stack.
    justify_items : str, optional
        Set in the style attribute of the flexbox containing this stack.
    align_content : str, optional
        Set in the style attribute of the flexbox containing this stack.
    align_items : str, optional
        Set in the style attribute of the flexbox containing this stack.

    Returns
    -------
    IPython.display.HTML
        Stacked elements which can be directly displayed in jupyter.
    '''
    jc = f'justify-content:{justify_content};'
    ji = f'justify-items:{justify_items};'
    ac = f'align-content:{align_content};'
    ai = f'align-items:{align_items};'

    raw_html = f'<div style="display:flex;flex-direction:row;{jc}{ji}{ac}{ai}">'
    for elem in args:
        raw_html += _to_html(elem, margin)
    raw_html += '</div>'
    return HTML(raw_html)

def vstack(*args, margin='5px 10px 5px 10px',
        justify_content='start', justify_items='start',
        align_content='start', align_items='start'):
    '''
    Vertically stack the html, markdown or string representation of `args`.

    The `args` will be stacked in order from top to bottom, and converted like this:

    * if the arg has the attribute `__repr__`, then its output is used pre-formatted,
    * if the arg has the attribute `_repr_html_`, then its output is used,
    * if the arg is a string, then it is treated as markdown and converted to HTML,
    * otherwise, the arg's default string conversion is used.

    Parameters
    ----------
    args
        Elements to stack vertically.

    margin : str, optional
        Set in the style attribute of a div wrapping elements in this stack
        (excluding lines, which have a fixed 5px margin). Give values in the
        order, eg. "5px 10px 5px 10px" (top right bottom left).
    justify_content, justify_items, align_content, align_items : str
        Set in the style attribute of the flexbox containing this stack.

    Returns
    -------
    IPython.display.HTML
        Stacked elements which can be directly displayed in jupyter.
    '''
    jc = f'justify-content:{justify_content};'
    ji = f'justify-items:{justify_items};'
    ac = f'align-content:{align_content};'
    ai = f'align-items:{align_items};'

    raw_html = f'<div style="display:flex;flex-direction:column;{jc}{ji}{ac}{ai}">'
    for elem in args:
        raw_html += _to_html(elem, margin)
    raw_html += '</div>'
    return HTML(raw_html)

def _to_html(arg, margin):
    fmt = IPython.get_ipython().display_formatter.formatters['text/html']
    def div(elem):
        return f'<div style="margin:{margin};">' + elem + '</div>'

    if isinstance(arg, Line):
        if arg is Line.HORIZONTAL:
            return f'<div style="border-top:solid 1px #AAAAAA;margin:5px;align-self:stretch;"></div>'
        elif arg is Line.VERTICAL:
            return f'<div style="border-left:solid 1px #AAAAAA;margin:5px;align-self:stretch;"></div>'

    if type(arg) is str:
        return div(markdown.markdown(arg, extensions=['tables']))

    if fmt is not None:
        result = fmt(arg)
        if result is None:
            result = '<pre>' + repr(arg) + '</pre>'
        return div(result)

    if (fmt is None) or (result is None):
        if hasattr(arg, '_repr_html_'):
            return div(arg._repr_html_())
        elif hasattr(arg, '__repr__'):
            raw = '<pre>' + repr(arg) + '</pre>'
            return div(raw)
        else:
            return div(str(arg))
    else:
        return result



"""
================================================
Process and capability (:mod:`mqr.plot.process`)
================================================

.. currentmodule:: mqr.plot.process

.. rubric:: Functions
.. autosummary::
    :toctree: generated/

    summary
    pdf
    tolerance
    capability
"""

import mqr
from mqr.process import Sample, Specification, Capability, Summary
from mqr.plot.defaults import Defaults
from mqr.plot.lib.util import set_kws

from matplotlib import pyplot as plt
import matplotlib.transforms as transforms
import numpy as np
import scipy.stats as st
import seaborn as sns

def pdf(sample: Sample, ax,
        nsigma=None, cp=None, bins='auto', show_long_term=False,
        short_kws=None, long_kws=None):
    """
    Plots a Gaussian PDF for the given sample.

    The PDF is truncated at six sigmas on each side.

    Parameters
    ----------
    sample : mqr.process.Sample
        Sample to plot.
    ax : matplotlib.axes.Axes
        Axes for the plot.
    nsigma : float, optional
        How many stddevs of the Gaussian PDF to plot (total). Default 6.
    cp : float, optional
        Width of the plot in triples of the standard deviation. Ie. when `cp` is
        2, the width will be 6 standard deviations. Default 2.
    bins : str, optional
        Passed to :func:`numpy.histogram_bin_edges`. Can be used to ensure the
        density corresponds to a histogram on the same plot.
    show_long_term : bool, optional
        Plots two more densities shifted left and right by 1.5 standard deviations.
    short_kws : dict, optional
        Keyword arguments passed to `matplotlib.pyplot.plot` for short-term
        densities.
    long_kws : dict, optional
        Keyword arguments passed to `matplotlib.pyplot.fill_between` for
        long-term densities.

    Examples
    --------
    See example in :mod:`mqr.plot.process`.
    """
    if (nsigma is not None) and (cp is not None):
        raise ValueError(f'Only one of `nsigma` or `cp` can be specified.')
    elif (nsigma is None) and (cp is None):
        nsigma = 6.0

    short_kws = set_kws(
        short_kws,
        color='C1',
        marker=2,
        markersize=8,
        mew=2,
        zorder=1,
    )
    long_kws = set_kws(
        long_kws,
        color='C2',
        alpha=0.1,
        zorder=1,
    )

    if cp is not None:
        nsigma = cp * 3

    dist = st.norm(sample.mean, sample.std)
    xmin_st = sample.mean - nsigma * sample.std
    xmax_st = sample.mean + nsigma * sample.std
    xs = np.linspace(xmin_st, xmax_st, 250)
    ys = mqr.plot.lib.util.scaled_density(xs, sample.data, bins=bins, dist=dist)

    line_kws = {k:v for k, v in short_kws.items() if k != 'marker'}
    ax.plot(xs, ys, **line_kws, label='Fitted density')
    ax.plot(xs[0], ys[0], **short_kws, label=f'$\\pm {nsigma:.2f} \\sigma$; $c_p={nsigma/3:.2f}$')
    ax.plot(xs[-1], ys[-1], **short_kws)

    if show_long_term:
        fill_kws = {k:v for k, v in long_kws.items() if k not in []}
        shift = 1.5 * sample.std
        dist_l = st.norm(sample.mean - shift, sample.std)
        dist_r = st.norm(sample.mean + shift, sample.std)
        ys_l = mqr.plot.lib.util.scaled_density(xs, sample.data, bins=bins, dist=dist_l)
        ys_r = mqr.plot.lib.util.scaled_density(xs, sample.data, bins=bins, dist=dist_r)
        ax.fill_between(xs, ys_l, **fill_kws, label='Long-term densities')
        ax.fill_between(xs, ys_r, **fill_kws)

def tolerance(spec: Specification, ax,
              prec=3, line_kws=None, tol_kws=None):
    """
    Plots tolerance region.

    Parameters
    ----------
    spec : mqr.process.Specification
        Spec containing tolerance bounds.
    ax : matplotlib.axes.Axes
        Axes for the plot.
    prec : int, optional
        Significant figures for the limit bounds.
    line_kws : dict, optional
        Keyword arguments for vertical lines at the centreline and limits. Passed
        to `matplotlib.pyplot.axvline`.
    tol_kws : dict, optional
        Keyword arguments for the shading over the tolerance region. Passed
        to `matplotlib.pyplot.axvspan`.

    Examples
    --------
    See example in :mod:`mqr.plot.process`.
    """
    line_kws = set_kws(
        line_kws,
        color='gray',
        zorder=-1,
    )
    tol_kws = set_kws(
        tol_kws,
        color='lightgray',
        zorder=-2,
    )

    ax.axvline(spec.target, **line_kws)
    ax.axvline(spec.lsl, **line_kws)
    ax.axvline(spec.usl, **line_kws)
    ax.axvspan(spec.lsl, spec.usl, **tol_kws, label='Tolerance')

    sec = ax.secondary_xaxis(location='top')
    sec.set_xticks([spec.lsl, spec.target, spec.usl])
    sec.tick_params(axis='x', which='major', direction='out',
                    top=True, labeltop=True, bottom=False, labelbottom=False)

def capability(summary: Summary, name: str, ax, nsigma=None, cp=None, show_long_term=False):
    """
    Plots all three of the process histogram, fitted normal distribution, and
    tolerance region for the sample called `name` in `summary`.

    Parameters
    ----------
    summary : mqr.process.Summary
        Summary containing sample with `name`.
    name : str
        Name of process whose capability will be shown.
    ax : matplotlib.axes.Axes
        Axes for the plot.
    nsigma : float, optional
        How many stddevs of the Gaussian PDF to plot (total). Default 6.
    cp : float, optional
        Width of the plot in triples of the standard deviation. Ie. when `cp` is
        2, the width will be 6 standard deviations. Default 2.
    show_long_term : bool, optional
        Plots two more densities shifted left and right by 1.5 standard deviations.

    Examples
    --------
    See example in :mod:`mqr.plot.process`.
    """
    if name not in summary.capabilities:
        raise ValueError(f'Summary has no specification/capability defined for {name}.')

    sample = summary[name]
    capability = summary.capabilities[name]
    specification = capability.spec

    sns.histplot(
        sample.data,
        color='C0',
        alpha=0.8,
        zorder=0,
        ax=ax,
        label='Sample histogram')
    tolerance(specification, ax=ax)
    pdf(sample=sample,
        show_long_term=show_long_term,
        nsigma=nsigma,
        cp=cp,
        ax=ax)
    ax.set
    ax.set_xlabel(
        f'{sample.name} ('
        f'target={specification.target:#.3g}, '
        f'cp={capability.cp:#.3g}, '
        f'cpk={capability.cpk:#.3g}'
        ')')
    ax.set_ylabel('count')

def summary(sample: Sample, ax, hyp_mean=None,
            hist_kws=None, box_kws=None, conf_kws=None):
    '''
    Histogram, boxplot and confidence interval for a sample.

    Best plotted on axes arranged vertically.

    Parameters
    ----------
    sample : mqr.process.Sample
        Sample to use for all three plots.
    ax : matplotlib.axes.Axes
        Axes for plot.
    hyp_mean: float, optional
        Hypothesised mean to plot on a confidence interval.

    Examples
    --------
    See example in :mod:`mqr.plot.process`.
    '''
    hist_kws = set_kws(
        hist_kws,
        color='C0',
    )
    box_kws = set_kws(
        box_kws,
        orient='h',
        color='C0',
    )
    conf_kws = set_kws(
        conf_kws,
    )

    ax = ax.flatten()
    if ax.shape != (3,):
        raise ValueError(f'Axes shape must be (3,) but is {ax.shape}.')

    sns.histplot(sample.data, ax=ax[0], **hist_kws)
    sns.boxplot(sample.data, ax=ax[1], **box_kws)
    mqr.plot.confint(sample.conf_mean, ax=ax[2], hyp_value=hyp_mean, **conf_kws)

    ax[1].sharex(ax[0])
    ax[2].sharex(ax[0])
    plt.setp(ax[0].get_xticklabels(), visible=False)
    plt.setp(ax[1].get_xticklabels(), visible=False)
    ax[1].tick_params(axis='y', left=False)
    ax[2].set_title('')
    ax[0].set_xlabel('')
    ax[1].set_xlabel('')
    ax[2].set_xlabel(sample.name)



import pytest

from mqr.inference.lib import util

def test_bounded_error_msg():
    assert (util.bounded_error_msg('abcdef') ==
        'Invalid bound "abcdef". Use both, below or above.')

def test_compare_error_msg():
    assert (util.compare_error_msg('abcdef') ==
        'Invalid comparison "abcdef". Use diff or ratio.')

def test_method_error_msg():
    assert (util.method_error_msg('abcdef', ['123x']) ==
        'Method "abcdef" is not available. Use 123x.')
    assert (util.method_error_msg('abcdef', ['123x', '123y']) ==
        'Method "abcdef" is not available. Use 123x or 123y.')
    assert (util.method_error_msg('abcdef', ['123x', '123y', '123z']) ==
        'Method "abcdef" is not available. Use 123x, 123y or 123z.')

def test_alternative_error_msg():
    assert (util.alternative_error_msg('abcdef') ==
        'Invalid alternative "abcdef". Use two-sided, less or greater.')



'''
Check call-throughs.
'''

import numbers
import numpy as np
import pytest

import mqr

def test_power_1sample():
    ra = 0.8
    hyp_rate = 0.5
    alpha = 0.05
    meas = 2.0
    nobs = 45

    alternative = 'two-sided'
    method = 'norm-approx'
    res = mqr.inference.rate.power_1sample(ra, hyp_rate, nobs, alpha, meas=meas, alternative=alternative, method=method)
    assert res.name == 'rate of events'
    assert res.alpha == alpha
    assert res.beta == pytest.approx(0.0512867)
    assert res.effect == '0.8 / 0.5 = 1.6'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_size == nobs

    alternative = 'greater'
    res = mqr.inference.rate.power_1sample(ra, hyp_rate, nobs, alpha, meas=meas, alternative=alternative, method=method)
    assert res.name == 'rate of events'
    assert res.alpha == alpha
    assert res.beta == pytest.approx(0.0299445)
    assert res.effect == '0.8 / 0.5 = 1.6'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_size == nobs

    ra = 0.3
    alternative = 'less'
    res = mqr.inference.rate.power_1sample(ra, hyp_rate, nobs, alpha, meas=meas, alternative=alternative, method=method)
    assert res.name == 'rate of events'
    assert res.alpha == alpha
    assert res.beta == pytest.approx(0.0900244)
    assert res.effect == '0.3 / 0.5 = 0.6'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_size == nobs

def test_power_2sample():
    r1 = 0.5
    r2 = 0.4
    nobs = 100
    alpha = 0.02

    compare = 'diff'
    H0_value = 0.0
    res = mqr.inference.rate.power_2sample(r1, r2, nobs, alpha, H0_value, compare=compare)
    assert res.name == 'difference between rates of events'
    assert res.alpha == alpha
    assert isinstance(res.beta, numbers.Number)
    assert res.effect == '0.5 - 0.4 = 0.1'
    assert res.alternative == 'two-sided'
    assert res.method == 'score'
    assert res.sample_size == nobs

    H0_value = None
    res = mqr.inference.rate.power_2sample(r1, r2, nobs, alpha, H0_value, compare=compare)
    assert res.name == 'difference between rates of events'
    assert res.alpha == alpha
    assert isinstance(res.beta, numbers.Number)
    assert res.effect == '0.5 - 0.4 = 0.1'
    assert res.alternative == 'two-sided'
    assert res.method == 'score'
    assert res.sample_size == nobs

    compare = 'ratio'
    H0_value = 1.0
    res = mqr.inference.rate.power_2sample(r1, r2, nobs, alpha, H0_value, compare=compare)
    assert res.name == 'ratio of rates of events'
    assert res.alpha == alpha
    assert isinstance(res.beta, numbers.Number)
    assert res.effect == '0.5 / 0.4 = 1.25'
    assert res.alternative == 'two-sided'
    assert res.method == 'score'
    assert res.sample_size == nobs

    H0_value = None
    res = mqr.inference.rate.power_2sample(r1, r2, nobs, alpha, H0_value, compare=compare)
    assert res.name == 'ratio of rates of events'
    assert res.alpha == alpha
    assert isinstance(res.beta, numbers.Number)
    assert res.effect == '0.5 / 0.4 = 1.25'
    assert res.alternative == 'two-sided'
    assert res.method == 'score'
    assert res.sample_size == nobs

def test_size_1sample():
    ra = 0.8
    hyp_rate = 0.5
    alpha = 0.05
    beta = 0.20

    alternative = 'two-sided'
    method = 'chi2'
    res = mqr.inference.rate.size_1sample(ra, hyp_rate, alpha, beta, alternative=alternative, method=method)
    assert res.name == 'rate of events'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == '0.8 / 0.5 = 1.6'
    assert res.alternative == alternative
    assert res.method == method
    assert np.ceil(res.sample_size) == 55

    alternative = 'greater'
    res = mqr.inference.rate.size_1sample(ra, hyp_rate, alpha, beta, alternative=alternative, method=method)
    assert res.name == 'rate of events'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == '0.8 / 0.5 = 1.6'
    assert res.alternative == alternative
    assert res.method == method
    assert np.ceil(res.sample_size) == 43

    ra = 0.3
    alternative = 'less'
    res = mqr.inference.rate.size_1sample(ra, hyp_rate, alpha, beta, alternative=alternative, method=method)
    assert res.name == 'rate of events'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == '0.3 / 0.5 = 0.6'
    assert res.alternative == alternative
    assert res.method == method
    assert np.ceil(res.sample_size) == 59

    ra = 0.8
    alternative = 'two-sided'
    method = 'norm-approx'
    res = mqr.inference.rate.size_1sample(ra, hyp_rate, alpha, beta, alternative=alternative, method=method)
    assert res.name == 'rate of events'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == '0.8 / 0.5 = 1.6'
    assert res.alternative == alternative
    assert res.method == method
    assert np.ceil(res.sample_size) == 51

    alternative = 'greater'
    res = mqr.inference.rate.size_1sample(ra, hyp_rate, alpha, beta, alternative=alternative, method=method)
    assert res.name == 'rate of events'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == '0.8 / 0.5 = 1.6'
    assert res.alternative == alternative
    assert res.method == method
    assert np.ceil(res.sample_size) == 41

    ra = 0.3
    alternative = 'less'
    res = mqr.inference.rate.size_1sample(ra, hyp_rate, alpha, beta, alternative=alternative, method=method)
    assert res.name == 'rate of events'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == '0.3 / 0.5 = 0.6'
    assert res.alternative == alternative
    assert res.method == method
    assert np.ceil(res.sample_size) == 66

def test_size_2sample():
    r1 = 0.8
    r2 = 0.5
    alpha = 0.05
    beta = 0.20
    effect = 0.2

    alternative = 'two-sided'
    res = mqr.inference.rate.size_2sample(r1, r2, alpha, beta, effect, alternative)
    assert res.name == 'difference between rates of events'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == '0.8 - 0.5 = 0.3'
    assert res.alternative == alternative
    assert res.method == 'score'
    assert isinstance(res.sample_size, numbers.Number)

    alternative = 'two-sided'
    res = mqr.inference.rate.size_2sample(r1, r2, alpha, beta, effect, alternative)
    assert res.name == 'difference between rates of events'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == '0.8 - 0.5 = 0.3'
    assert res.alternative == alternative
    assert res.method == 'score'
    assert isinstance(res.sample_size, numbers.Number)

    effect = 0.0
    method = 'z'
    alternative = 'less'
    res = mqr.inference.rate.size_2sample(r1, r2, alpha, beta, effect, alternative, method=method)
    assert res.name == 'difference between rates of events'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == '0.8 - 0.5 = 0.3'
    assert res.alternative == alternative
    assert res.method == method
    assert isinstance(res.sample_size, numbers.Number)

    alternative = 'two-sided'
    res = mqr.inference.rate.size_2sample(r1, r2, alpha, beta, effect, alternative, method=method)
    assert res.name == 'difference between rates of events'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == '0.8 - 0.5 = 0.3'
    assert res.alternative == alternative
    assert res.method == method
    assert isinstance(res.sample_size, numbers.Number)

    effect = 0.2
    method = 'score'
    res = mqr.inference.rate.size_2sample(r1, r2, alpha, beta, effect, alternative, method=method)
    assert res.name == 'difference between rates of events'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == '0.8 - 0.5 = 0.3'
    assert res.alternative == alternative
    assert res.method == 'score'
    assert isinstance(res.sample_size, numbers.Number)

    res = mqr.inference.rate.size_2sample(r1, r2, alpha, beta, effect, alternative, method=method)
    assert res.name == 'difference between rates of events'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == '0.8 - 0.5 = 0.3'
    assert res.alternative == alternative
    assert res.method == 'score'
    assert isinstance(res.sample_size, numbers.Number)

    compare = 'diff'
    res = mqr.inference.rate.size_2sample(r1, r2, alpha, beta, None, alternative, compare=compare)
    assert res.name == 'difference between rates of events'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == '0.8 - 0.5 = 0.3'
    assert res.alternative == alternative
    assert res.method == 'score'
    assert isinstance(res.sample_size, numbers.Number)

    compare = 'ratio'
    res = mqr.inference.rate.size_2sample(r1, r2, alpha, beta, None, alternative, compare=compare)
    assert res.name == 'ratio of rates of events'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == '0.8 / 0.5 = 1.6'
    assert res.alternative == alternative
    assert res.method == 'score'
    assert isinstance(res.sample_size, numbers.Number)

def test_confint_1sample():
    count = 5
    n = 20
    meas = 2.0
    conf = 0.90
    bounded = 'both'
    method = 'score'

    res = mqr.inference.rate.confint_1sample(count, n, meas, conf, bounded, method=method)
    assert res.name == 'rate of events'
    assert res.method == 'score'
    assert res.value == count / n / meas
    assert res.conf == conf
    assert res.bounded == bounded
    assert isinstance(res.lower, numbers.Number)
    assert isinstance(res.upper, numbers.Number)

    count = 10
    n = 10
    meas = 2.0
    conf = 0.95

    bounded = 'both'
    method = 'chi2'
    res = mqr.inference.rate.confint_1sample(count, n, meas, conf, bounded, method)
    assert list(res) == pytest.approx([0.2398, 0.9195], abs=1e-4)
    method = 'exact'
    res = mqr.inference.rate.confint_1sample(count, n, meas, conf, bounded, method)
    assert list(res) == pytest.approx([0.2746, 0.9195], abs=1e-4)
    method = 'wald-cc'
    res = mqr.inference.rate.confint_1sample(count, n, meas, conf, bounded, method)
    assert list(res) == pytest.approx([0.1729, 0.8426], abs=1e-4)

    bounded = 'below'
    method = 'chi2'
    res = mqr.inference.rate.confint_1sample(count, n, meas, conf, bounded, method)
    assert list(res) == pytest.approx([0.2713, np.inf], abs=1e-4)
    method = 'exact'
    res = mqr.inference.rate.confint_1sample(count, n, meas, conf, bounded, method)
    assert list(res) == pytest.approx([0.3085, np.inf], abs=1e-4)
    method = 'wald-cc'
    res = mqr.inference.rate.confint_1sample(count, n, meas, conf, bounded, method)
    assert list(res) == pytest.approx([0.2215, np.inf], abs=1e-4)

    bounded = 'above'
    method = 'chi2'
    res = mqr.inference.rate.confint_1sample(count, n, meas, conf, bounded, method)
    assert list(res) == pytest.approx([0.0, 0.8481], abs=1e-4)
    method = 'exact'
    res = mqr.inference.rate.confint_1sample(count, n, meas, conf, bounded, method)
    assert list(res) == pytest.approx([0.0, 0.8481], abs=1e-4)
    method = 'wald-cc'
    res = mqr.inference.rate.confint_1sample(count, n, meas, conf, bounded, method)
    assert list(res) == pytest.approx([0.0, 0.7915], abs=1e-4)

def test_confint_2sample():
    count1 = 5
    n1 = 20
    meas1 = 2.0
    count2 = 8
    n2 = 18
    meas2 = 3
    conf = 0.90
    bounded = 'both'
    method = 'score'

    res = mqr.inference.rate.confint_2sample(count1, n1, count2, n2, meas1, meas2, conf, method='score')
    assert res.name == 'difference between rates of events'
    assert res.method == 'score'
    assert res.value == count1 / n1 / meas1 - count2 / n2 / meas2
    assert isinstance(res.lower, numbers.Number)
    assert isinstance(res.upper, numbers.Number)
    assert res.conf == conf
    assert res.bounded == bounded

    count1, n1, meas1 = 1234, 20, 2
    count2, n2, meas2 = 2345, 20, 1
    conf = 0.95

    bounded = 'both'
    method = 'wald'
    res = mqr.inference.rate.confint_2sample(count1, n1, count2, n2, meas1, meas2, conf, bounded=bounded, method=method)
    assert list(res) == pytest.approx([-91.4481, -81.3519], abs=1e-4)
    method = 'wald-moment'
    res = mqr.inference.rate.confint_2sample(count1, n1, count2, n2, meas1, meas2, conf, bounded=bounded, method=method)
    assert list(res) == pytest.approx([-91.4963, -81.3997], abs=1e-4)

    bounded = 'below'
    method = 'wald'
    res = mqr.inference.rate.confint_2sample(count1, n1, count2, n2, meas1, meas2, conf, bounded=bounded, method=method)
    assert list(res) == pytest.approx([-90.6365, np.inf], abs=1e-4)
    method = 'wald-moment'
    res = mqr.inference.rate.confint_2sample(count1, n1, count2, n2, meas1, meas2, conf, bounded=bounded, method=method)
    assert list(res) == pytest.approx([-90.6705, np.inf], abs=1e-4)

    bounded = 'above'
    method = 'wald'
    res = mqr.inference.rate.confint_2sample(count1, n1, count2, n2, meas1, meas2, conf, bounded=bounded, method=method)
    assert list(res) == pytest.approx([-np.inf, -82.1635], abs=1e-4)
    method = 'wald-moment'
    res = mqr.inference.rate.confint_2sample(count1, n1, count2, n2, meas1, meas2, conf, bounded=bounded, method=method)
    assert list(res) == pytest.approx([-np.inf, -82.1972], abs=1e-4)

def test_confint_2sample_krishnamoorthy_lee():
    """
    Checks calculations against the values in Table 7 in [1]_.

    References
    ----------
    .. [1] Krishnamoorthy, K., & Lee, M. (2013).
       New approximate confidence intervals for the difference between
       two Poisson means and comparison.
       Journal of Statistical Computation and Simulation, 83(12), 2232-2243.
    """
    lower, upper = mqr.inference.lib.rate.confint_2sample_wald(
        3, 310, 7, 3500,
        1.0, 1.0,
        0.95, 'both')
    assert lower == pytest.approx(-0.0034, abs=1e-4)
    assert upper == pytest.approx(0.0187, abs=1e-4)

    lower, upper = mqr.inference.lib.rate.confint_2sample_wald(
        3, 310, 7, 3500,
        1.0, 1.0,
        0.95, 'below')
    assert lower == pytest.approx(-0.00160, abs=1e-5)
    assert upper == np.inf

    lower, upper = mqr.inference.lib.rate.confint_2sample_wald(
        3, 310, 7, 3500,
        1.0, 1.0,
        0.95, 'above')
    assert lower == -np.inf
    assert upper == pytest.approx(0.01695, abs=1e-5)

    lower, upper = mqr.inference.lib.rate.confint_2sample_wald_moment(
        3, 310, 7, 3500,
        1.0, 1.0,
        0.95, 'both')
    assert lower == pytest.approx(0.0009, abs=1e-4)
    assert upper == pytest.approx(0.02573, abs=1e-5)

    lower, upper = mqr.inference.lib.rate.confint_2sample_wald_moment(
        3, 310, 7, 3500,
        1.0, 1.0,
        0.95, 'below')
    assert lower == pytest.approx(0.00156, abs=1e-5)
    assert upper == np.inf

    lower, upper = mqr.inference.lib.rate.confint_2sample_wald_moment(
        3, 310, 7, 3500,
        1.0, 1.0,
        0.95, 'above')
    assert lower == -np.inf
    assert upper == pytest.approx(0.02175, abs=1e-5)

def test_test_1sample():
    count = 20
    n = 30
    meas = 3
    H0_rate = 0.3
    alternative = 'two-sided'
    method = 'exact-c'

    res = mqr.inference.rate.test_1sample(count, n, meas, H0_rate, alternative)
    assert res.description == 'rate of events'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'rate'
    assert res.sample_stat_target == H0_rate
    assert res.sample_stat_value == pytest.approx(count / n / meas)
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

def test_test_2sample():
    count1 = 3
    n1 = 10
    meas1 = 6
    count2 = 4
    n2 = 15
    meas2 = 5
    H0_value = -0.1
    alternative = 'less'
    method = 'wald'

    compare = 'diff'
    res = mqr.inference.rate.test_2sample(count1, n1, count2, n2, meas1, meas2, H0_value, alternative, method, compare)
    assert res.description == 'difference between rates of events'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'rate1 - rate2'
    assert res.sample_stat_target == H0_value
    assert res.sample_stat_value == pytest.approx(count1 / n1 / meas1 - count2 / n2 / meas2)
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    compare = 'ratio'
    res = mqr.inference.rate.test_2sample(count1, n1, count2, n2, meas1, meas2, H0_value, alternative, method, compare)
    assert res.description == 'ratio of rates of events'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'rate1 / rate2'
    assert res.sample_stat_target == H0_value
    assert res.sample_stat_value == pytest.approx(count1 / n1 / meas1 / (count2 / n2 / meas2))
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    compare = 'diff'
    res = mqr.inference.rate.test_2sample(count1, n1, count2, n2, meas1, meas2, None, alternative, method, compare)
    assert res.description == 'difference between rates of events'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'rate1 - rate2'
    assert res.sample_stat_target == 0.0
    assert res.sample_stat_value == pytest.approx(count1 / n1 / meas1 - (count2 / n2 / meas2))
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    compare = 'ratio'
    res = mqr.inference.rate.test_2sample(count1, n1, count2, n2, meas1, meas2, None, alternative, method, compare)
    assert res.description == 'ratio of rates of events'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'rate1 / rate2'
    assert res.sample_stat_target == 1.0
    assert res.sample_stat_value == pytest.approx(count1 / n1 / meas1 / (count2 / n2 / meas2))
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)



"""
==================================================
Basic statistical inference (:mod:`mqr.inference`)
==================================================

.. :currentmodule:`mqr.inference`

User guide
    :doc:`/user_guide/inference`

Detailed examples
    https://github.com/nklsxn/mqrpy-guide

This package contains wrappers and functions for basic inference on statistics
calculated from samples of data. Many routines call either :mod:`scipy` or
:mod:`statsmodels` (see function docs).

One of the main goals of this module is to organise statistical tests by
purpose. The name of a test or method used in a routine is given in the
``method`` argument, instead of the function name.

Routines in the top level (:mod:`mqr.inference`) are parametric, while
routines in the :mod:`mqr.inference.nonparametric` module are non-parametric.

The modules implement hypothesis tests for all of the listed statistics,
and in addition, the parametric modules also calculate confidence intervals and
sample-size.

.. rubric:: Parametric statistics
.. autosummary::
    :template: autosummary/submodule.rst
    :toctree: generated/

    dist
    correlation
    mean
    proportion
    rate
    stddev
    variance

.. rubric:: Non-parametric statistics
.. autosummary::
    :template: autosummary/submodule.rst

    nonparametric.correlation
    nonparametric.dist
    nonparametric.median
    nonparametric.quantile
    nonparametric.variance

.. rubric:: Result types
.. autosummary::
    :toctree: generated/

    ~confint.ConfidenceInterval
    ~hyptest.HypothesisTest
    ~power.TestPower

Notes
-----
NIST [1]_ has further information on how to compare processes and products using
hypothesis tests. 

Examples
--------
All functions have a similar interface and return one of the result types above.

This example calculates a confidence interval for the difference between Pearson
correlation coefficients.

>>> x = np.array([1, 2, 1, 1, 2])
>>> y = np.array([3, 1, 2, 2, 1])
>>> mqr.inference.correlation.confint(x, y)
ConfidenceInterval(
    name='correlation',
    method='fisher-z',
    value=-0.8728715609439694,
    lower=-0.9915444093872248,
    upper=0.040865942695342376,
    conf=0.95,
    bounded='both')

This example calculates the sample size required to achieve an hypothesis test
on the mean with standardised effect size (Cohen's F) of 0.5, 95% confidence and
90% power.

>>> import mqr
>>> mqr.inference.mean.size_1sample(0.5, 0.05, 0.1, 'greater')
TestPower(
    name='mean',
    alpha=0.05,
    beta=0.1,
    effect=0.5,
    alternative='greater',
    method='t',
    sample_size=35.65267806948644)

Here is an example of an hypothesis test on the ratio of rates of events. Two
samples were taken, one before a process improvement activity and another after.
The sample before showed 5 defects in a sample of fabric 0.5m*0.5m, which is 20
defects per squqre metre. After the improvement, the same sized sample showed
only 1 defect --- that's 4 defects per square metre on average. This tests the
hypothesis that the defect rate has been decreased by at least 10 defects per
square metre.

>>> import mqr
>>> mqr.inference.rate.test_2sample(
>>>     5, 0.25, 1, 0.25, H0_value=10,
>>>     alternative='less', compare='diff')
HypothesisTest(
    description='difference between rates of events',
    alternative='less',
    method='score',
    sample_stat='rate1 - rate2',
    sample_stat_target=10,
    sample_stat_value=16.0,
    stat=0.65209454185206,
    pvalue=0.7428299075363056,
    null='rate1 - rate2 == 10',
    alt='rate1 - rate2 < 10')

From these samples, there is not enough evidence to conclude that the defect
rate has been improved by 10 defects per square metre.

This is an example of a test for normality. First, create a random sample of
F-distributed data.

>>> import numpy as np
>>> import scipy.stats as st
>>> np.random.seed(0)
>>> data = st.f(1, 2).rvs(40)

Check the hypothesis that they are normally distributed using the Kolmogorov-
Smirnov test.

>>> import mqr
>>> mqr.inference.dist.test_1sample(data, test='ks-norm')
HypothesisTest(
    description='non-normality',
    alternative='two-sided',
    method='kolmogorov-smirnov',
    sample_stat='distribution',
    sample_stat_target='normal',
    sample_stat_value=None,
    stat=0.2881073311024548,
    pvalue=0.0009999999999998899,
    null='distribution == normal',
    alt='distribution != normal')

References
----------
.. [1] National Institute of Standards and Technology (US Dep. of Commerce).
       "Product and Process Comparisons."
       https://www.itl.nist.gov/div898/handbook/prc/prc.htm
"""

import mqr.inference.nonparametric

import mqr.inference.correlation
import mqr.inference.dist
import mqr.inference.mean
import mqr.inference.power
import mqr.inference.proportion
import mqr.inference.rate
import mqr.inference.stddev
import mqr.inference.variance



from mergedeep import merge
import matplotlib.transforms as transforms
import numpy as np
import scipy

from mqr.plot.defaults import Defaults

def grouped_df(data, ax,
               line_kws=None, text_kws=None):
    """
    Plots from a pandas.DataFrame with columns grouped along the x-axis.

    Parameters
    ----------
    data : pandas.DataFrame
        Dataframe whose columns will be plot next to each other.
    ax : matplotlib.axes.Axes
        Axes for plot.
    line_kws : dict
        Keyword args for lines, passed to :func:`matplotlib..plot <matplotlib.pyplot.plot>`.
    text_kws : dict
        Keyword args for text, passed to :func:`matplotlib..text <matplotlib.pyplot.text>`.
    """
    line_kws = set_kws(
        line_kws,
        color='C0',
        marker=Defaults.marker,
        markersize=5,
    )
    text_kws = set_kws(
        text_kws,
        rotation=0,
        rotation_mode='anchor',
        va='top',
        ha='left',
    )

    idx_labels = data.index
    idx = np.arange(len(idx_labels))
    M = len(idx)

    groups = data.columns
    N = len(groups)

    tr = transforms.blended_transform_factory(ax.transData, ax.transAxes)
    for i, g in enumerate(groups):
        xs = idx + i * M
        ax.plot(xs, data.loc[:, g], label=g, **line_kws)
        sep_x = i * M - 0.5
        ax.axvline(sep_x, linewidth=0.8, color='gray', linestyle=(0, (5, 5)))
        ax.text(sep_x+0.25, 0.99, g, transform=tr, **text_kws)

    ax.set_xticks(np.arange(len(idx)*N))
    ax.set_xticklabels(np.tile(idx_labels, N), rotation=90)

def scaled_density(xs, data, dist=None, bins='auto'):
    """
    Calculate the PDF of the points `xs`, scaled to match a histogram.

    Parameters
    ----------
    xs : array_like
        X values.
    data : array_like
        Sampled random data.
    dist : :class:`scipy.stats.rv_continuous`
        Distribution of the density.
    bins : str, optional
        Passed to :func:`numpy.histogram_bin_edges` to calculate bin widths.

    Returns
    -------
    array_like
        Y values corresponding to `xs` for the scaled density.
    """
    if dist is None:
        dist = scipy.stats.norm.fit(data)
    elif isinstance(dist, scipy.stats.distributions.rv_continuous):
        dist = dist(*dist.fit(data))
    N = len(data)
    edges = np.histogram_bin_edges(data, bins=bins)
    binwidth = edges[1] - edges[0]
    ys = dist.pdf(xs)
    return ys * N * binwidth

def set_kws(input_kws, **default_kws):
    """
    Merges dicts of keyword arguments.

    Parameters
    ----------
    input_kws : dict
        Overriding keyword arguments that, when specified, replace the values
        in `default_kws`.
    default_kws : dict
        Default values that can be overriden by `input_kws`.

    Returns
    -------
    dict
        A set of keyword arguments containing all of `input_kws` and any of
        `default_kws` whose key did not appear in `default_kws`.
    """
    if input_kws is None:
        input_kws = {}
    return merge(default_kws, input_kws)



import mqr

from collections.abc import Iterable
import IPython.core.formatters
import pandas as pd

class HTMLFormatter(IPython.core.formatters.HTMLFormatter):
    def __call__(self, obj):
        if isinstance(obj, mqr.process.Sample):
            return format_samples((obj,))

        elif isinstance(obj, mqr.process.Summary):
            return format_samples(obj.samples.values())

        elif is_iterable_of(obj, mqr.process.Sample):
            return format_samples(obj)

        elif is_dict_of(obj, mqr.process.Sample):
            return format_samples(obj.values())

        elif isinstance(obj, mqr.process.Capability):
            return format_capabilities((obj,))

        elif is_iterable_of(obj, mqr.process.Capability):
            return format_capabilities(obj)

        elif is_dict_of(obj, mqr.process.Capability):
            return format_capabilities(obj.values())

        else:
            return super().__call__(obj)

def is_iterable_of(obj, typ):
    return (
        isinstance(obj, Iterable) and
        all([isinstance(elem, typ) for elem in obj])
    )

def is_dict_of(obj, typ):
    return (
        isinstance(obj, dict) and
        all([isinstance(elem, typ) for elem in obj.values()])
    )

def format_samples(samples):
    display_fmt = mqr.notebook.Defaults.sample_value_fmt

    def join(s):
        return ''.join(s)

    def th(scope='col'):
        def _th(s):
            return f'<th scope="{scope}">{s}</th>'
        return _th

    def td(s):
        return f'<td>{s}</td>'

    def fmt_value(value):
        return f'{value:{display_fmt}}'
    
    col_headers = [s.name for s in samples]

    ad_stat = [s.ad_stat for s in samples]
    ad_pvalue = [s.ad_pvalue for s in samples]
    ks_stat = [s.ks_stat for s in samples]
    ks_pvalue = [s.ks_pvalue for s in samples]

    nobs = [s.nobs for s in samples]
    mean = [s.mean for s in samples]
    std = [s.std for s in samples]
    var = [s.var for s in samples]
    skewness = [s.skewness for s in samples]
    kurtosis = [s.kurtosis for s in samples]
    minimum = [s.minimum for s in samples]
    quartile1 = [s.quartile1 for s in samples]
    median = [s.median for s in samples]
    quartile3 = [s.quartile3 for s in samples]
    maximum = [s.maximum for s in samples]

    outliers = [len(s.outliers) for s in samples]

    html = f'''
    <table>
        <thead>
            <tr>
                <th scope="col"></th>
                {join(map(th(), col_headers))}
            </tr>
        </thead>
        <tbody>
            <tr>
                <th colspan={len(col_headers)+1} style="text-align:left;">Normality (Anderson-Darling)</th>
            </tr>
            <tr>
                <th scope="row">Stat</th>
                {join(map(td, map(fmt_value, ad_stat)))}
            </tr>
            <tr>
                <th scope="row">P-value</th>
                {join(map(td, map(fmt_value, ad_pvalue)))}
            </tr>

            <thead><tr></tr></thead>
            <tr>
                <th scope="row">N</th>
                {join(map(td, nobs))}
            </tr>

            <thead><tr></tr></thead>
            <tr>
                <th scope="row">Mean</th>
                {join(map(td, map(fmt_value, mean)))}
            </tr>
            <tr>
                <th scope="row">StdDev</th>
                {join(map(td, map(fmt_value, std)))}
            </tr>
            <tr>
                <th scope="row">Variance</th>
                {join(map(td, map(fmt_value, var)))}
            </tr>
            <tr>
                <th scope="row">Skewness</th>
                {join(map(td, map(fmt_value, skewness)))}
            </tr>
            <tr>
                <th scope="row">Kurtosis</th>
                {join(map(td, map(fmt_value, kurtosis)))}
            </tr>

            <thead><tr></tr></thead>
            <tr>
                <th scope="row">Minimum</th>
                {join(map(td, map(fmt_value, minimum)))}
            </tr>
            <tr>
                <th scope="row">1st Quartile</th>
                {join(map(td, map(fmt_value, quartile1)))}
            </tr>
            <tr>
                <th scope="row">Median</th>
                {join(map(td, map(fmt_value, median)))}
            </tr>
            <tr>
                <th scope="row">3rd Quartile</th>
                {join(map(td, map(fmt_value, quartile3)))}
            </tr>
            <tr>
                <th scope="row">Maximum</th>
                {join(map(td, map(fmt_value, maximum)))}
            </tr>

            <thead><tr></tr></thead>
            <tr>
                <th scope="row">N Outliers</th>
                {join(map(td, outliers))}
            </tr>
        </tbody>
        <tfoot>
        </tfoot>
    </table>
    '''
    return html

def format_capabilities(capabilities):
    display_fmt = mqr.notebook.Defaults.capability_value_fmt

    def join(s):
        return ''.join(s)

    def th(scope='col'):
        def _th(s):
            return f'<th scope="{scope}">{s}</th>'
        return _th

    def td(s):
        return f'<td>{s}</td>'

    def bold(s):
        return f'<b>{s}</b>'

    def gray(s):
        return f'<font color="gray">{s}</font>'

    def fmt(value):
        return f'{value:{display_fmt}}'

    def compose(f, g):
        return lambda *a, **kw: f(g(*a, **kw))

    names = [c.sample.name for c in capabilities]
    specs = [c.spec for c in capabilities]

    return f'''
    <table>
        <thead>
            <tr>
                <th scope="col"></th>
                {join([th()(n) for n in names])}
            </tr>
        </thead>
        <tbody>
            <tr>
                <th scope="row"><font color="gray">USL</font></th>
                {join([td(gray(fmt(s.usl))) for s in specs])}
            </tr
            <tr>
                <th scope="row">Target</th>
                {join([td(fmt(s.target)) for s in specs])}
            </tr>
            <tr>
                <th scope="row"><font color="gray">LSL</font></th>
                {join([td(gray(fmt(s.lsl))) for s in specs])}
            </tr>

            <thead><tr></tr></thead>
            <tr>
                <th scope="row"><b>C<sub>pk</sub></b></th>
                {join([td(bold(fmt(s.cpk))) for s in capabilities])}
            </tr>
            <tr>
                <th scope="row">C<sub>p</sub></th>
                {join([td(fmt(s.cp)) for s in capabilities])}
            </tr>
            <tr>
                <th scope="row">Defects<sub>st</sub> (ppm)</th>
                {join([td(fmt(s.defects_st*1e6)) for s in capabilities])}
            </tr>
            <tr>
                <th scope="row">Defects<sub>lt</sub> (ppm)</th>
                {join([td(fmt(s.defects_lt*1e6)) for s in capabilities])}
            </tr>
        <tbody>
        <tfoot>
        </tfoot>
    </table>
    '''



from mqr.inference.hyptest import HypothesisTest

import mqr.inference.lib.util as util

import scipy
import statsmodels

def test_1sample(x, method='runs', cutoff='median'):
    """
    Hypothesis test on randomness characteristics of a sample.

    Null-hypotheses by method
        | 'runs'
        |   observations in `x` are independent and identically distributed
        | 'adf'
        |   observations in `x` are non-stationary (time series has a unit root)

    Parameters
    ----------
    x : array_like
        Test the hypothesis that `x` is random.
    method : {'runs'}, optional
        | 'runs'
        |   Runs test for independent and identical distribution.
            Calls :func:`sm..runstest_1samp <statsmodels.sandbox.stats.runs.runstest_1samp>`.
    cutoff : str, optional
        The cutoff to group large and small values. Only used for method "runs".

    Returns
    -------
    :class:`mqr.inference.hyptest.HypothesisTest`
    """
    if method == 'runs':
        description = 'randomness'
        sample_stat_target = 'iid'
        stat, pvalue = statsmodels.sandbox.stats.runs.runstest_1samp(x, cutoff=cutoff, correction=True)
    else:
        raise ValueError(util.method_error_msg(method, ['runs']))

    return HypothesisTest(
        description=description,
        alternative='two-sided',
        method=method,
        sample_stat='dist(x)',
        sample_stat_target=sample_stat_target,
        sample_stat_value=None,
        stat=stat,
        pvalue=pvalue,)

def test_2sample(x, y, alternative='two-sided', method='ks'):
    """
    Hypothesis test for distributions of two samples.

    Null-hypothesis
        cdf(`x`) == cdf(`y`)

    Parameters
    ----------
    x, y : array_like
        Samples to compare.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : {'ks', 'runs'}, optional
        | 'ks'
        |   Kolmogorov-Smirnov test.
            Calls :func:`scipy..kstest <scipy.stats.kstest>`.
        | 'runs'
        |   Wald-Wolfowitz test
            Calls :func:`sm..runstest_2samp <statsmodels.sandbox.stats.runs.runstest_2samp>`.

    Returns
    -------
    :class:`mqr.inference.hyptest.HypothesisTest`
    """
    if method == 'ks':
        method = 'kolmogorov-smirnov'
        stat, pvalue = scipy.stats.ks_2samp(
            data1=x,
            data2=y,
            alternative=alternative)
    elif method == 'runs':
        if alternative != 'two-sided':
            raise ValueError('Only "two-sided" alternative available.')
        stat, pvalue = statsmodels.sandbox.stats.runs.runstest_2samp(x, y, correction=True)
    else:
        raise ValueError(util.method_error_msg(method, ['ks', 'runs']))

    return HypothesisTest(
        description='sampling distribution',
        alternative=alternative,
        method=method,
        sample_stat='dist(x)',
        sample_stat_target='dist(y)',
        sample_stat_value=None,
        stat=stat,
        pvalue=pvalue)
